from awxkit.utils import random_title


class HasSurvey(object):
    def add_survey(self, spec=None, name=None, description=None, required=False, enabled=True):
        payload = dict(name=name or 'Survey - {}'.format(random_title()),
                       description=description or random_title(10),
                       spec=spec or [dict(required=required,
                                          question_name="What's the password?",
                                          variable="secret",
                                          type="password",
                                          default="foo")])
        if enabled != self.survey_enabled:
            self.patch(survey_enabled=enabled)
        return self.related.survey_spec.post(payload).get()



import json

from awxkit.utils import (
    filter_by_class,
    not_provided,
    random_title,
    suppress,
    update_payload,
    PseudoNamespace)
from awxkit.api.pages import Credential, Inventory, Project, UnifiedJobTemplate
from awxkit.api.mixins import HasCreate, HasInstanceGroups, HasNotifications, HasSurvey, HasCopy, DSAdapter
from awxkit.api.resources import resources
import awxkit.exceptions as exc
from . import base
from . import page


class JobTemplate(
        HasCopy,
        HasCreate,
        HasInstanceGroups,
        HasNotifications,
        HasSurvey,
        UnifiedJobTemplate):

    optional_dependencies = [Inventory, Credential, Project]
    NATURAL_KEY = ('organization', 'name')

    def launch(self, payload={}):
        """Launch the job_template using related->launch endpoint."""
        # get related->launch
        launch_pg = self.get_related('launch')

        # launch the job_template
        result = launch_pg.post(payload)

        # return job
        if result.json['type'] == 'job':
            jobs_pg = self.get_related('jobs', id=result.json['job'])
            assert jobs_pg.count == 1, \
                "job_template launched (id:%s) but job not found in response at %s/jobs/" % \
                (result.json['job'], self.url)
            return jobs_pg.results[0]
        elif result.json['type'] == 'workflow_job':
            slice_workflow_jobs = self.get_related(
                'slice_workflow_jobs', id=result.json['id'])
            assert slice_workflow_jobs.count == 1, (
                "job_template launched sliced job (id:%s) but not found in related %s/slice_workflow_jobs/" %
                (result.json['id'], self.url)
            )
            return slice_workflow_jobs.results[0]
        else:
            raise RuntimeError('Unexpected type of job template spawned job.')

    def payload(self, job_type='run', playbook='ping.yml', **kwargs):
        name = kwargs.get('name') or 'JobTemplate - {}'.format(random_title())
        description = kwargs.get('description') or random_title(10)
        payload = PseudoNamespace(
            name=name,
            description=description,
            job_type=job_type)

        optional_fields = (
            'ask_scm_branch_on_launch',
            'ask_credential_on_launch',
            'ask_diff_mode_on_launch',
            'ask_inventory_on_launch',
            'ask_job_type_on_launch',
            'ask_limit_on_launch',
            'ask_skip_tags_on_launch',
            'ask_tags_on_launch',
            'ask_variables_on_launch',
            'ask_verbosity_on_launch',
            'allow_simultaneous',
            'become_enabled',
            'diff_mode',
            'force_handlers',
            'forks',
            'host_config_key',
            'job_tags',
            'limit',
            'skip_tags',
            'start_at_task',
            'survey_enabled',
            'timeout',
            'use_fact_cache',
            'vault_credential',
            'verbosity',
            'job_slice_count',
            'webhook_service',
            'webhook_credential',
            'scm_branch')

        update_payload(payload, optional_fields, kwargs)

        extra_vars = kwargs.get('extra_vars', not_provided)
        if extra_vars != not_provided:
            if isinstance(extra_vars, dict):
                extra_vars = json.dumps(extra_vars)
            payload.update(extra_vars=extra_vars)

        if kwargs.get('project'):
            payload.update(project=kwargs.get('project').id, playbook=playbook)
        if kwargs.get('inventory'):
            payload.update(inventory=kwargs.get('inventory').id)
        if kwargs.get('credential'):
            payload.update(credential=kwargs.get('credential').id)
        if kwargs.get('webhook_credential'):
            webhook_cred = kwargs.get('webhook_credential')
            if isinstance(webhook_cred, int):
                payload.update(webhook_credential=int(webhook_cred))
            elif hasattr(webhook_cred, 'id'):
                payload.update(webhook_credential=webhook_cred.id)
            else:
                raise AttributeError("Webhook credential must either be integer of pkid or Credential object")

        return payload

    def add_label(self, label):
        if isinstance(label, page.Page):
            label = label.json
        with suppress(exc.NoContent):
            self.related.labels.post(label)

    def create_payload(
            self,
            name='',
            description='',
            job_type='run',
            playbook='ping.yml',
            credential=Credential,
            inventory=Inventory,
            project=None,
            **kwargs):
        if not project:
            project = Project
        if not inventory and not kwargs.get('ask_inventory_on_launch', False):
            inventory = Inventory

        self.create_and_update_dependencies(
            *
            filter_by_class(
                (credential,
                 Credential),
                (inventory,
                 Inventory),
                (project,
                 Project)))
        project = self.ds.project if project else None
        inventory = self.ds.inventory if inventory else None
        credential = self.ds.credential if credential else None

        payload = self.payload(
            name=name,
            description=description,
            job_type=job_type,
            playbook=playbook,
            credential=credential,
            inventory=inventory,
            project=project,
            **kwargs)
        payload.ds = DSAdapter(self.__class__.__name__, self._dependency_store)
        return payload, credential

    def create(
            self,
            name='',
            description='',
            job_type='run',
            playbook='ping.yml',
            credential=Credential,
            inventory=Inventory,
            project=None,
            **kwargs):
        payload, credential = self.create_payload(name=name, description=description, job_type=job_type,
                                                  playbook=playbook, credential=credential, inventory=inventory,
                                                  project=project, **kwargs)
        ret = self.update_identity(
            JobTemplates(
                self.connection).post(payload))
        if credential:
            with suppress(exc.NoContent):
                self.related.credentials.post(dict(id=credential.id))
        if 'vault_credential' in kwargs:
            with suppress(exc.NoContent):
                if not isinstance(kwargs['vault_credential'], int):
                    raise ValueError(
                        "Expected 'vault_credential' value to be an integer, the id of the desired vault credential")
                self.related.credentials.post(
                    dict(id=kwargs['vault_credential']))
        return ret

    def add_credential(self, credential):
        with suppress(exc.NoContent):
            self.related.credentials.post(
                dict(id=credential.id, associate=True))

    def remove_credential(self, credential):
        with suppress(exc.NoContent):
            self.related.credentials.post(
                dict(id=credential.id, disassociate=True))

    def remove_all_credentials(self):
        for cred in self.related.credentials.get().results:
            with suppress(exc.NoContent):
                self.related.credentials.post(
                    dict(id=cred.id, disassociate=True))


page.register_page([resources.job_template,
                    (resources.job_templates, 'post'),
                    (resources.job_template_copy, 'post')], JobTemplate)


class JobTemplates(page.PageList, JobTemplate):

    pass


page.register_page([resources.job_templates,
                    resources.related_job_templates], JobTemplates)


class JobTemplateCallback(base.Base):

    pass


page.register_page(resources.job_template_callback, JobTemplateCallback)


class JobTemplateLaunch(base.Base):

    pass


page.register_page(resources.job_template_launch, JobTemplateLaunch)


class JobTemplateCopy(base.Base):

    pass


page.register_page([resources.job_template_copy], JobTemplateCopy)



from awxkit.api.mixins import HasCreate, HasCopy, DSAdapter
from awxkit.api.pages import Organization
from awxkit.api.resources import resources
from awxkit.config import config
import awxkit.exceptions as exc
from awxkit.utils import not_provided, random_title, suppress, PseudoNamespace
from . import base
from . import page


job_results = ('any', 'error', 'success')
notification_types = (
    'email',
    'irc',
    'pagerduty',
    'slack',
    'twilio',
    'webhook',
    'mattermost',
    'grafana',
    'rocketchat')


class NotificationTemplate(HasCopy, HasCreate, base.Base):

    dependencies = [Organization]
    NATURAL_KEY = ('organization', 'name')

    def test(self):
        """Create test notification"""
        assert 'test' in self.related, \
            "No such related attribute 'test'"

        # trigger test notification
        notification_id = self.related.test.post().notification

        # return notification page
        notifications_pg = self.get_related(
            'notifications', id=notification_id).wait_until_count(1)
        assert notifications_pg.count == 1, \
            "test notification triggered (id:%s) but notification not found in response at %s/notifications/" % \
            (notification_id, self.url)
        return notifications_pg.results[0]

    def silent_delete(self):
        """Delete the Notification Template, ignoring the exception that is raised
        if there are notifications pending.
        """
        try:
            super(NotificationTemplate, self).silent_delete()
        except (exc.MethodNotAllowed):
            pass

    def payload(self, organization, notification_type='slack', messages=not_provided, **kwargs):
        payload = PseudoNamespace(
            name=kwargs.get('name') or 'NotificationTemplate ({0}) - {1}' .format(
                notification_type,
                random_title()),
            description=kwargs.get('description') or random_title(10),
            organization=organization.id,
            notification_type=notification_type)
        if messages != not_provided:
            payload['messages'] = messages

        notification_configuration = kwargs.get(
            'notification_configuration', {})
        payload.notification_configuration = notification_configuration

        if payload.notification_configuration == {}:
            services = config.credentials.notification_services

            if notification_type == 'email':
                fields = (
                    'host',
                    'username',
                    'password',
                    'port',
                    'use_ssl',
                    'use_tls',
                    'sender',
                    'recipients')
                cred = services.email
            elif notification_type == 'irc':
                fields = (
                    'server',
                    'port',
                    'use_ssl',
                    'password',
                    'nickname',
                    'targets')
                cred = services.irc
            elif notification_type == 'pagerduty':
                fields = ('client_name', 'service_key', 'subdomain', 'token')
                cred = services.pagerduty
            elif notification_type == 'slack':
                fields = ('channels', 'token')
                cred = services.slack
            elif notification_type == 'twilio':
                fields = (
                    'account_sid',
                    'account_token',
                    'from_number',
                    'to_numbers')
                cred = services.twilio
            elif notification_type == 'webhook':
                fields = ('url', 'headers')
                cred = services.webhook
            elif notification_type == 'mattermost':
                fields = (
                    'mattermost_url',
                    'mattermost_username',
                    'mattermost_channel',
                    'mattermost_icon_url',
                    'mattermost_no_verify_ssl')
                cred = services.mattermost
            elif notification_type == 'grafana':
                fields = ('grafana_url',
                          'grafana_key')
                cred = services.grafana
            elif notification_type == 'rocketchat':
                fields = ('rocketchat_url',
                          'rocketchat_no_verify_ssl')
                cred = services.rocketchat
            else:
                raise ValueError(
                    'Unknown notification_type {0}'.format(notification_type))

            for field in fields:
                if field == 'bot_token':
                    payload_field = 'token'
                else:
                    payload_field = field
                value = kwargs.get(field, cred.get(field, not_provided))
                if value != not_provided:
                    payload.notification_configuration[payload_field] = value

        return payload

    def create_payload(
            self,
            name='',
            description='',
            notification_type='slack',
            organization=Organization,
            messages=not_provided,
            **kwargs):
        if notification_type not in notification_types:
            raise ValueError(
                'Unsupported notification type "{0}".  Please use one of {1}.' .format(
                    notification_type, notification_types))
        self.create_and_update_dependencies(organization)
        payload = self.payload(
            organization=self.ds.organization,
            notification_type=notification_type,
            name=name,
            description=description,
            messages=messages,
            **kwargs)
        payload.ds = DSAdapter(self.__class__.__name__, self._dependency_store)
        return payload

    def create(
            self,
            name='',
            description='',
            notification_type='slack',
            organization=Organization,
            messages=not_provided,
            **kwargs):
        payload = self.create_payload(
            name=name,
            description=description,
            notification_type=notification_type,
            organization=organization,
            messages=messages,
            **kwargs)
        return self.update_identity(
            NotificationTemplates(
                self.connection).post(payload))

    def associate(self, resource, job_result='any'):
        """Associates a NotificationTemplate with the provided resource"""
        return self._associate(resource, job_result)

    def disassociate(self, resource, job_result='any'):
        """Disassociates a NotificationTemplate with the provided resource"""
        return self._associate(resource, job_result, disassociate=True)

    def _associate(self, resource, job_result='any', disassociate=False):
        if job_result not in job_results:
            raise ValueError(
                'Unsupported job_result type "{0}".  Please use one of {1}.' .format(
                    job_result, job_results))

        result_attr = 'notification_templates_{0}'.format(job_result)
        if result_attr not in resource.related:
            raise ValueError(
                'Unsupported resource "{0}".  Does not have a related {1} field.' .format(
                    resource, result_attr))

        payload = dict(id=self.id)
        if disassociate:
            payload['disassociate'] = True

        with suppress(exc.NoContent):
            getattr(resource.related, result_attr).post(payload)


page.register_page([resources.notification_template,
                    (resources.notification_templates, 'post'),
                    (resources.notification_template_copy, 'post'),
                    resources.notification_template_any,
                    resources.notification_template_started,
                    resources.notification_template_error,
                    resources.notification_template_success,
                    resources.notification_template_approval], NotificationTemplate)


class NotificationTemplates(page.PageList, NotificationTemplate):

    pass


page.register_page([resources.notification_templates,
                    resources.related_notification_templates,
                    resources.notification_templates_any,
                    resources.notification_templates_started,
                    resources.notification_templates_error,
                    resources.notification_templates_success,
                    resources.notification_templates_approvals],
                   NotificationTemplates)


class NotificationTemplateCopy(base.Base):

    pass


page.register_page(resources.notification_template_copy, NotificationTemplateCopy)


class NotificationTemplateTest(base.Base):

    pass


page.register_page(
    resources.notification_template_test,
    NotificationTemplateTest)



from awxkit.api.resources import resources
from . import base
from . import page


class CredentialInputSource(base.Base):

    pass


page.register_page(resources.credential_input_source, CredentialInputSource)


class CredentialInputSources(page.PageList, CredentialInputSource):

    pass


page.register_page([resources.credential_input_sources,
                    resources.related_input_sources], CredentialInputSources)



from contextlib import contextmanager
from datetime import datetime, timedelta, tzinfo
import inspect
import logging
import random
import shlex
import types
import time
import sys
import re
import os

import yaml

from awxkit.words import words
from awxkit.exceptions import WaitUntilTimeout


log = logging.getLogger(__name__)


cloud_types = (
    'aws',
    'azure',
    'azure_ad',
    'azure_classic',
    'azure_rm',
    'cloudforms',
    'ec2',
    'gce',
    'openstack',
    'openstack_v2',
    'openstack_v3',
    'rhv',
    'satellite6',
    'tower',
    'vmware')
credential_type_kinds = ('cloud', 'net')

not_provided = 'xx__NOT_PROVIDED__xx'


def super_dir_set(cls):
    attrs = set()
    for _class in inspect.getmro(cls):
        attrs.update(dir(_class))
    return attrs


class NoReloadError(Exception):
    pass


class PseudoNamespace(dict):

    def __init__(self, _d=None, **loaded):
        if not isinstance(_d, dict):
            _d = {}
        _d.update(loaded)
        super(PseudoNamespace, self).__init__(_d)

        # Convert nested structures into PseudoNamespaces
        for k, v in _d.items():
            tuple_converted = False
            if isinstance(v, tuple):
                self[k] = v = list(v)
                tuple_converted = True

            if isinstance(v, list):
                for i, item in enumerate(v):
                    if isinstance(item, dict):
                        self[k][i] = PseudoNamespace(item)
                if tuple_converted:
                    self[k] = tuple(self[k])
            elif isinstance(v, dict):
                self[k] = PseudoNamespace(v)

    def __getattr__(self, attr):
        try:
            return self.__getitem__(attr)
        except KeyError:
            raise AttributeError(
                "{!r} has no attribute {!r}".format(
                    self.__class__.__name__, attr))

    def __setattr__(self, attr, value):
        self.__setitem__(attr, value)

    def __setitem__(self, key, value):
        if not isinstance(value, PseudoNamespace):
            tuple_converted = False
            if isinstance(value, dict):
                value = PseudoNamespace(value)
            elif isinstance(value, tuple):
                value = list(value)
                tuple_converted = True

            if isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, dict) and not isinstance(item, PseudoNamespace):
                        value[i] = PseudoNamespace(item)
                if tuple_converted:
                    value = tuple(value)

        super(PseudoNamespace, self).__setitem__(key, value)

    def __delattr__(self, attr):
        self.__delitem__(attr)

    def __dir__(self):
        attrs = super_dir_set(self.__class__)
        attrs.update(self.keys())
        return sorted(attrs)

    # override builtin in order to have updated content become
    # PseudoNamespaces if applicable
    def update(self, iterable=None, **kw):
        if iterable:
            if (hasattr(iterable,
                        'keys') and isinstance(iterable.keys,
                                               (types.FunctionType,
                                                types.BuiltinFunctionType,
                                                types.MethodType))):
                for key in iterable:
                    self[key] = iterable[key]
            else:
                for (k, v) in iterable:
                    self[k] = v
        for k in kw:
            self[k] = kw[k]


def is_relative_endpoint(candidate):
    return isinstance(candidate, (str,)) and candidate.startswith('/api/')


def is_class_or_instance(obj, cls):
    """returns True is obj is an instance of cls or is cls itself"""
    return isinstance(obj, cls) or obj is cls


def filter_by_class(*item_class_tuples):
    """takes an arbitrary number of (item, class) tuples and returns a list consisting
    of each item if it's an instance of the class, the item if it's a (class, dict()) tuple,
    the class itself if item is truthy but not an instance of the
    class or (class, dict()) tuple, or None if item is falsy in the same order as the arguments
    ```
    _cred = Credential()
    inv, org, cred = filter_base_subclasses((True, Inventory), (None, Organization), (_cred, Credential))
    inv == Inventory
    org == None
    cred == _cred
    ```
    """
    results = []
    for item, cls in item_class_tuples:
        if item:
            was_tuple = False
            if isinstance(item, tuple):
                was_tuple = True
                examined_item = item[0]
            else:
                examined_item = item
            if is_class_or_instance(
                    examined_item,
                    cls) or is_proper_subclass(
                    examined_item,
                    cls):
                results.append(item)
            else:
                updated = (cls, item[1]) if was_tuple else cls
                results.append(updated)
        else:
            results.append(None)
    return results


def load_credentials(filename=None):
    if filename is None:
        path = os.path.join(os.getcwd(), 'credentials.yaml')
    else:
        path = os.path.abspath(filename)

    if os.path.isfile(path):
        with open(path) as credentials_fh:
            credentials_dict = yaml.safe_load(credentials_fh)
            return credentials_dict
    else:
        msg = 'Unable to load credentials file at %s' % path
        raise Exception(msg)


def load_projects(filename=None):
    if filename is None:
        return {}
    else:
        path = os.path.abspath(filename)

    if os.path.isfile(path):
        with open(path) as projects_fh:
            projects_dict = yaml.safe_load(projects_fh)
            return projects_dict
    else:
        msg = 'Unable to load projects file at %s' % path
        raise Exception(msg)


def logged_sleep(duration, level='DEBUG', stack_depth=1):
    level = getattr(logging, level.upper())
    # based on
    # http://stackoverflow.com/questions/1095543/get-name-of-calling-functions-module-in-python
    try:
        frm = inspect.stack()[stack_depth]
        logger = logging.getLogger(inspect.getmodule(frm[0]).__name__)
    except AttributeError:  # module is None (interactive shell)
        logger = log  # fall back to utils logger
    logger.log(level, 'Sleeping for {0} seconds.'.format(duration))
    time.sleep(duration)


def poll_until(function, interval=5, timeout=0):
    """Polls `function` every `interval` seconds until it returns a non-falsey
    value. If this does not occur within the provided `timeout`,
    a WaitUntilTimeout is raised.

    Each attempt will log the time that has elapsed since the original
    request.
    """
    start_time = time.time()

    while True:
        elapsed = time.time() - start_time
        log.debug('elapsed: {0:4.1f}'.format(elapsed))

        value = function()
        if value:
            return value

        if elapsed > timeout:
            break

        logged_sleep(interval, stack_depth=3)

    msg = 'Timeout after {0} seconds.'.format(elapsed)
    raise WaitUntilTimeout(None, msg)


def gen_utf_char():
    is_char = False
    b = 'b'
    while not is_char:
        b = random.randint(32, 0x10ffff)
        is_char = chr(b).isprintable()
    return chr(b)


def random_int(maxint=sys.maxsize):
    max = int(maxint)
    return random.randint(0, max)


def random_ipv4():
    """Generates a random ipv4 address;; useful for testing."""
    return ".".join(str(random.randint(1, 255)) for i in range(4))


def random_ipv6():
    """Generates a random ipv6 address;; useful for testing."""
    return ':'.join(
        '{0:x}'.format(
            random.randint(
                0,
                2 ** 16 -
                1)) for i in range(8))


def random_loopback_ip():
    """Generates a random loopback ipv4 address;; useful for testing."""
    return "127.{}.{}.{}".format(
        random_int(255),
        random_int(255),
        random_int(255))


def random_utf8(*args, **kwargs):
    """This function exists due to a bug in ChromeDriver that throws an
    exception when a character outside of the BMP is sent to `send_keys`.
    Code pulled from http://stackoverflow.com/a/3220210.
    """
    pattern = re.compile('[^\u0000-\uD7FF\uE000-\uFFFF]', re.UNICODE)
    length = args[0] if len(args) else kwargs.get('length', 10)
    scrubbed = pattern.sub('\uFFFD', ''.join(
        [gen_utf_char() for _ in range(length)]))

    return scrubbed


def random_title(num_words=2, non_ascii=True):
    base = ''.join([random.choice(words) for word in range(num_words)])
    if non_ascii:
        title = ''.join([base, random_utf8(1)])
    else:
        title = ''.join([base, ''.join([str(random_int()) for _ in range(3)])])
    return title


def update_payload(payload, fields, kwargs):
    """Takes a list of fields and adds their kwargs value to payload if defined.
    If the payload has an existing value and not_provided is the kwarg value for that key,
    the existing key/value are stripped from the payload.
    """
    not_provided_as_kwarg = 'xx_UPDATE_PAYLOAD_FIELD_NOT_PROVIDED_AS_KWARG_xx'
    for field in fields:
        field_val = kwargs.get(field, not_provided_as_kwarg)
        if field_val not in (not_provided, not_provided_as_kwarg):
            payload[field] = field_val
        elif field in payload and field_val == not_provided:
            payload.pop(field)
    return payload


def to_str(obj):
    if isinstance(obj, bytes):
        return obj.decode('utf-8')
    return obj


def to_bool(obj):
    if isinstance(obj, (str,)):
        return obj.lower() not in ('false', 'off', 'no', 'n', '0', '')
    return bool(obj)


def load_json_or_yaml(obj):
    try:
        return yaml.safe_load(obj)
    except AttributeError:
        raise TypeError("Provide valid YAML/JSON.")


def get_class_if_instance(obj):
    if not inspect.isclass(obj):
        return obj.__class__
    return obj


def class_name_to_kw_arg(class_name):
    """'ClassName' -> 'class_name'"""
    first_pass = re.sub(r'([a-z])([A-Z0-9])', r'\1_\2', class_name)
    second_pass = re.sub(r'([0-9])([a-zA-Z])', r'\1_\2', first_pass).lower()
    return second_pass.replace('v2_', '')


def is_proper_subclass(obj, cls):
    return inspect.isclass(obj) and obj is not cls and issubclass(obj, cls)


def are_same_endpoint(first, second):
    """Equivalence check of two urls, stripped of query parameters"""
    def strip(url):
        return url.replace('www.', '').split('?')[0]
    return strip(first) == strip(second)


@contextmanager
def suppress(*exceptions):
    """Context manager that suppresses the provided exceptions

    :param exceptions: List of exceptions to suppress

    Usage::
        >>> with suppress(ZeroDivisionError):
        >>>     foo = 1/0
        >>>     # This code will not run

    Note: This is an intermediate framework and test refactoring tool.
    It's almost never a good idea to plan on using this. Also, note
    that after the suppressed exception has been caught, no further
    statements in the with block will be executed.
    """
    try:
        yield
    except exceptions:
        pass


def utcnow():
    """Provide a wrapped copy of the built-in utcnow that can be easily mocked."""
    return datetime.utcnow()


class UTC(tzinfo):
    """Concrete implementation of tzinfo for UTC. For more information, see:
    https://docs.python.org/2/library/datetime.html
    """

    def tzname(self, dt):
        return 'UTC'

    def dst(self, dt):
        return timedelta(0)

    def utcoffset(self, dt):
        return timedelta(0)


def seconds_since_date_string(
        date_str,
        fmt='%Y-%m-%dT%H:%M:%S.%fZ',
        default_tz=UTC()):
    """Return the number of seconds since the date and time indicated by a date
    string and its corresponding format string.

    :param date_str: string representing a date and time.
    :param fmt: Formatting string - by default, this value is set to parse
        date strings originating from awx API response data.
    :param default_tz: Assumed tzinfo if the parsed date_str does not include tzinfo

    For more information on python date string formatting directives, see
        https://docs.python.org/2/library/datetime.httpsml#strftime-strptime-behavior
    """
    parsed_datetime = datetime.strptime(date_str, fmt)

    if not parsed_datetime.tzinfo:
        parsed_datetime = parsed_datetime.replace(tzinfo=default_tz)

    elapsed = utcnow().replace(tzinfo=UTC()) - parsed_datetime

    return elapsed.total_seconds()


def to_ical(dt):
    return re.sub('[:-]', '', dt.strftime("%Y%m%dT%H%M%SZ"))


def version_from_endpoint(endpoint):
    return endpoint.split('/api/')[1].split('/')[0] or 'common'


def args_string_to_list(args):
    """Converts cmdline arg string to list of args.  The reverse of subprocess.list2cmdline()
    heavily inspired by robot.utils.argumentparser.cmdline2list()
    """
    lexer = shlex.shlex(args, posix=True)
    lexer.escapedquotes = '"\''
    lexer.commenters = ''
    lexer.whitespace_split = True
    return [token.decode('utf-8') for token in lexer]


def is_list_or_tuple(item):
    return isinstance(item, list) or isinstance(item, tuple)



import logging
import json
import re

from awxkit.api.pages import (
    Credential,
    Organization,
    Project,
    UnifiedJob,
    UnifiedJobTemplate
)
from awxkit.utils import (
    filter_by_class,
    random_title,
    update_payload,
    suppress,
    not_provided,
    PseudoNamespace,
    poll_until,
    random_utf8
)
from awxkit.api.mixins import DSAdapter, HasCreate, HasInstanceGroups, HasNotifications, HasVariables, HasCopy
from awxkit.api.resources import resources
import awxkit.exceptions as exc
from . import base
from . import page


log = logging.getLogger(__name__)


class Inventory(HasCopy, HasCreate, HasInstanceGroups, HasVariables, base.Base):

    dependencies = [Organization]
    NATURAL_KEY = ('organization', 'name')

    def print_ini(self):
        """Print an ini version of the inventory"""
        output = list()
        inv_dict = self.related.script.get(hostvars=1).json

        for group in inv_dict.keys():
            if group == '_meta':
                continue

            # output host groups
            output.append('[%s]' % group)
            for host in inv_dict[group].get('hosts', []):
                # FIXME ... include hostvars
                output.append(host)
            output.append('')  # newline

            # output child groups
            if inv_dict[group].get('children', []):
                output.append('[%s:children]' % group)
                for child in inv_dict[group].get('children', []):
                    output.append(child)
                output.append('')  # newline

            # output group vars
            if inv_dict[group].get('vars', {}).items():
                output.append('[%s:vars]' % group)
                for k, v in inv_dict[group].get('vars', {}).items():
                    output.append('%s=%s' % (k, v))
                output.append('')  # newline

        print('\n'.join(output))

    def payload(self, organization, **kwargs):
        payload = PseudoNamespace(
            name=kwargs.get('name') or 'Inventory - {}'.format(
                random_title()),
            description=kwargs.get('description') or random_title(10),
            organization=organization.id)

        optional_fields = (
            'host_filter',
            'insights_credential',
            'kind',
            'variables')

        update_payload(payload, optional_fields, kwargs)

        if 'variables' in payload and isinstance(payload.variables, dict):
            payload.variables = json.dumps(payload.variables)
        if 'insights_credential' in payload and isinstance(
                payload.insights_credential, Credential):
            payload.insights_credential = payload.insights_credential.id

        return payload

    def create_payload(
            self,
            name='',
            description='',
            organization=Organization,
            **kwargs):
        self.create_and_update_dependencies(organization)
        payload = self.payload(
            name=name,
            description=description,
            organization=self.ds.organization,
            **kwargs)
        payload.ds = DSAdapter(self.__class__.__name__, self._dependency_store)
        return payload

    def create(
            self,
            name='',
            description='',
            organization=Organization,
            **kwargs):
        payload = self.create_payload(
            name=name,
            description=description,
            organization=organization,
            **kwargs)
        return self.update_identity(
            Inventories(
                self.connection).post(payload))

    def add_host(self, host=None):
        if host is None:
            return self.related.hosts.create(inventory=self)

        if isinstance(host, base.Base):
            host = host.json
        with suppress(exc.NoContent):
            self.related.hosts.post(host)
        return host

    def wait_until_deleted(self):
        def _wait():
            try:
                self.get()
            except exc.NotFound:
                return True
        poll_until(_wait, interval=1, timeout=60)

    def update_inventory_sources(self, wait=False):
        response = self.related.update_inventory_sources.post()
        source_ids = [entry['inventory_source']
                      for entry in response if entry['status'] == 'started']

        inv_updates = []
        for source_id in source_ids:
            inv_source = self.related.inventory_sources.get(
                id=source_id).results.pop()
            inv_updates.append(inv_source.related.current_job.get())

        if wait:
            for update in inv_updates:
                update.wait_until_completed()
        return inv_updates


page.register_page([resources.inventory,
                    (resources.inventories, 'post'),
                    (resources.inventory_copy, 'post')], Inventory)


class Inventories(page.PageList, Inventory):

    pass


page.register_page([resources.inventories,
                    resources.related_inventories], Inventories)


class InventoryScript(HasCopy, HasCreate, base.Base):

    dependencies = [Organization]

    def payload(self, organization, **kwargs):
        payload = PseudoNamespace(
            name=kwargs.get('name') or 'Inventory Script - {}'.format(
                random_title()),
            description=kwargs.get('description') or random_title(10),
            organization=organization.id,
            script=kwargs.get('script') or self._generate_script())
        return payload

    def create_payload(
            self,
            name='',
            description='',
            organization=Organization,
            script='',
            **kwargs):
        self.create_and_update_dependencies(organization)
        payload = self.payload(
            name=name,
            description=description,
            organization=self.ds.organization,
            script=script,
            **kwargs)
        payload.ds = DSAdapter(self.__class__.__name__, self._dependency_store)
        return payload

    def create(
            self,
            name='',
            description='',
            organization=Organization,
            script='',
            **kwargs):
        payload = self.create_payload(
            name=name,
            description=description,
            organization=organization,
            script=script,
            **kwargs)
        return self.update_identity(
            InventoryScripts(
                self.connection).post(payload))

    def _generate_script(self):
        script = '\n'.join([
            '#!/usr/bin/env python',
            '# -*- coding: utf-8 -*-',
            'import json',
            'inventory = dict()',
            'inventory["{0}"] = dict()',
            'inventory["{0}"]["hosts"] = list()',
            'inventory["{0}"]["hosts"].append("{1}")',
            'inventory["{0}"]["hosts"].append("{2}")',
            'inventory["{0}"]["hosts"].append("{3}")',
            'inventory["{0}"]["hosts"].append("{4}")',
            'inventory["{0}"]["hosts"].append("{5}")',
            'inventory["{0}"]["vars"] = dict(ansible_host="127.0.0.1", ansible_connection="local")',
            'print(json.dumps(inventory))'
        ])
        group_name = re.sub(r"[\']", "", "group-{}".format(random_utf8()))
        host_names = [
            re.sub(
                r"[\':]",
                "",
                "host_{}".format(
                    random_utf8())) for _ in range(5)]

        return script.format(group_name, *host_names)


page.register_page([resources.inventory_script,
                    (resources.inventory_scripts, 'post'),
                    (resources.inventory_script_copy, 'post')], InventoryScript)


class InventoryScripts(page.PageList, InventoryScript):

    pass


page.register_page([resources.inventory_scripts], InventoryScripts)


class Group(HasCreate, HasVariables, base.Base):

    dependencies = [Inventory]
    optional_dependencies = [Credential, InventoryScript]

    @property
    def is_root_group(self):
        """Returns whether the current group is a top-level root group in the inventory"""
        return self.related.inventory.get().related.root_groups.get(id=self.id).count == 1

    def get_parents(self):
        """Inspects the API and returns all groups that include the current group as a child."""
        return Groups(self.connection).get(children=self.id).results

    def payload(self, inventory, credential=None, **kwargs):
        payload = PseudoNamespace(
            name=kwargs.get('name') or 'Group{}'.format(
                random_title(
                    non_ascii=False)),
            description=kwargs.get('description') or random_title(10),
            inventory=inventory.id)

        if credential:
            payload.credential = credential.id

        update_payload(payload, ('variables',), kwargs)

        if 'variables' in payload and isinstance(payload.variables, dict):
            payload.variables = json.dumps(payload.variables)

        return payload

    def create_payload(
            self,
            name='',
            description='',
            inventory=Inventory,
            credential=None,
            source_script=None,
            **kwargs):
        credential, source_script = filter_by_class(
            (credential, Credential), (source_script, InventoryScript))
        self.create_and_update_dependencies(
            inventory, credential, source_script)
        credential = self.ds.credential if credential else None
        payload = self.payload(
            inventory=self.ds.inventory,
            credential=credential,
            name=name,
            description=description,
            **kwargs)
        payload.ds = DSAdapter(self.__class__.__name__, self._dependency_store)
        return payload

    def create(self, name='', description='', inventory=Inventory, **kwargs):
        payload = self.create_payload(
            name=name,
            description=description,
            inventory=inventory,
            **kwargs)

        parent = kwargs.get('parent', None)  # parent must be a Group instance
        resource = parent.related.children if parent else Groups(
            self.connection)
        return self.update_identity(resource.post(payload))

    def add_host(self, host=None):
        if host is None:
            host = self.related.hosts.create(inventory=self.ds.inventory)
            with suppress(exc.NoContent):
                host.related.groups.post(dict(id=self.id))
            return host

        if isinstance(host, base.Base):
            host = host.json
        with suppress(exc.NoContent):
            self.related.hosts.post(host)
        return host

    def add_group(self, group):
        if isinstance(group, page.Page):
            group = group.json
        with suppress(exc.NoContent):
            self.related.children.post(group)

    def remove_group(self, group):
        if isinstance(group, page.Page):
            group = group.json
        with suppress(exc.NoContent):
            self.related.children.post(dict(id=group.id, disassociate=True))


page.register_page([resources.group,
                    (resources.groups, 'post')], Group)


class Groups(page.PageList, Group):

    pass


page.register_page([resources.groups,
                    resources.host_groups,
                    resources.inventory_related_groups,
                    resources.inventory_related_root_groups,
                    resources.group_children,
                    resources.group_potential_children], Groups)


class Host(HasCreate, HasVariables, base.Base):

    dependencies = [Inventory]

    def payload(self, inventory, **kwargs):
        payload = PseudoNamespace(
            name=kwargs.get('name') or 'Host{}'.format(
                random_title(
                    non_ascii=False)),
            description=kwargs.get('description') or random_title(10),
            inventory=inventory.id)

        optional_fields = ('enabled', 'instance_id')

        update_payload(payload, optional_fields, kwargs)

        variables = kwargs.get('variables', not_provided)

        if variables is None:
            variables = dict(
                ansible_host='127.0.0.1',
                ansible_connection='local')

        if variables != not_provided:
            if isinstance(variables, dict):
                variables = json.dumps(variables)
            payload.variables = variables

        return payload

    def create_payload(
            self,
            name='',
            description='',
            variables=None,
            inventory=Inventory,
            **kwargs):
        self.create_and_update_dependencies(
            *filter_by_class((inventory, Inventory)))
        payload = self.payload(
            inventory=self.ds.inventory,
            name=name,
            description=description,
            variables=variables,
            **kwargs)
        payload.ds = DSAdapter(self.__class__.__name__, self._dependency_store)
        return payload

    def create(
            self,
            name='',
            description='',
            variables=None,
            inventory=Inventory,
            **kwargs):
        payload = self.create_payload(
            name=name,
            description=description,
            variables=variables,
            inventory=inventory,
            **kwargs)
        return self.update_identity(Hosts(self.connection).post(payload))


page.register_page([resources.host,
                    (resources.hosts, 'post')], Host)


class Hosts(page.PageList, Host):

    pass


page.register_page([resources.hosts,
                    resources.group_related_hosts,
                    resources.inventory_related_hosts,
                    resources.inventory_sources_related_hosts], Hosts)


class FactVersion(base.Base):

    pass


page.register_page(resources.host_related_fact_version, FactVersion)


class FactVersions(page.PageList, FactVersion):

    @property
    def count(self):
        return len(self.results)


page.register_page(resources.host_related_fact_versions, FactVersions)


class FactView(base.Base):

    pass


page.register_page(resources.fact_view, FactView)


class InventorySource(HasCreate, HasNotifications, UnifiedJobTemplate):

    optional_schedule_fields = tuple()
    dependencies = [Inventory]
    optional_dependencies = [Credential, InventoryScript, Project]
    NATURAL_KEY = ('organization', 'name', 'inventory')

    def payload(
            self,
            inventory,
            source='custom',
            credential=None,
            source_script=None,
            project=None,
            **kwargs):
        payload = PseudoNamespace(
            name=kwargs.get('name') or 'InventorySource - {}'.format(
                random_title()),
            description=kwargs.get('description') or random_title(10),
            inventory=inventory.id,
            source=source)

        if credential:
            payload.credential = credential.id
        if source_script:
            payload.source_script = source_script.id
        if project:
            payload.source_project = project.id

        optional_fields = (
            'group_by',
            'instance_filters',
            'source_path',
            'source_regions',
            'source_vars',
            'timeout',
            'overwrite',
            'overwrite_vars',
            'update_cache_timeout',
            'update_on_launch',
            'update_on_project_update',
            'verbosity')

        update_payload(payload, optional_fields, kwargs)

        return payload

    def create_payload(
            self,
            name='',
            description='',
            source='custom',
            inventory=Inventory,
            credential=None,
            source_script=InventoryScript,
            project=None,
            **kwargs):
        if source != 'custom' and source_script == InventoryScript:
            source_script = None
        if source == 'scm':
            kwargs.setdefault('overwrite_vars', True)
            if project is None:
                project = Project

        inventory, credential, source_script, project = filter_by_class((inventory, Inventory),
                                                                        (credential, Credential),
                                                                        (source_script, InventoryScript),
                                                                        (project, Project))
        self.create_and_update_dependencies(
            inventory, credential, source_script, project)

        if credential:
            credential = self.ds.credential
        if source_script:
            source_script = self.ds.inventory_script
        if project:
            project = self.ds.project

        payload = self.payload(
            inventory=self.ds.inventory,
            source=source,
            credential=credential,
            source_script=source_script,
            project=project,
            name=name,
            description=description,
            **kwargs)
        payload.ds = DSAdapter(self.__class__.__name__, self._dependency_store)
        return payload

    def create(
            self,
            name='',
            description='',
            source='custom',
            inventory=Inventory,
            credential=None,
            source_script=InventoryScript,
            project=None,
            **kwargs):
        payload = self.create_payload(
            name=name,
            description=description,
            source=source,
            inventory=inventory,
            credential=credential,
            source_script=source_script,
            project=project,
            **kwargs)
        return self.update_identity(
            InventorySources(
                self.connection).post(payload))

    def update(self):
        """Update the inventory_source using related->update endpoint"""
        # get related->launch
        update_pg = self.get_related('update')

        # assert can_update == True
        assert update_pg.can_update, \
            "The specified inventory_source (id:%s) is not able to update (can_update:%s)" % \
            (self.id, update_pg.can_update)

        # start the inventory_update
        result = update_pg.post()

        # assert JSON response
        assert 'inventory_update' in result.json, \
            "Unexpected JSON response when starting an inventory_update.\n%s" % \
            json.dumps(result.json, indent=2)

        # locate and return the inventory_update
        jobs_pg = self.related.inventory_updates.get(
            id=result.json['inventory_update'])
        assert jobs_pg.count == 1, \
            "An inventory_update started (id:%s) but job not found in response at %s/inventory_updates/" % \
            (result.json['inventory_update'], self.url)
        return jobs_pg.results[0]

    @property
    def is_successful(self):
        """An inventory_source is considered successful when source != "" and super().is_successful ."""
        return self.source != "" and super(
            InventorySource, self).is_successful

    def add_credential(self, credential):
        with suppress(exc.NoContent):
            self.related.credentials.post(
                dict(id=credential.id, associate=True))

    def remove_credential(self, credential):
        with suppress(exc.NoContent):
            self.related.credentials.post(
                dict(id=credential.id, disassociate=True))


page.register_page([resources.inventory_source,
                    (resources.inventory_sources, 'post')], InventorySource)


class InventorySources(page.PageList, InventorySource):

    pass


page.register_page([resources.inventory_sources,
                    resources.related_inventory_sources],
                   InventorySources)


class InventorySourceGroups(page.PageList, Group):

    pass


page.register_page(
    resources.inventory_sources_related_groups,
    InventorySourceGroups)


class InventorySourceUpdate(base.Base):

    pass


page.register_page([resources.inventory_sources_related_update,
                    resources.inventory_related_update_inventory_sources],
                   InventorySourceUpdate)


class InventoryUpdate(UnifiedJob):

    pass


page.register_page(resources.inventory_update, InventoryUpdate)


class InventoryUpdates(page.PageList, InventoryUpdate):

    pass


page.register_page([resources.inventory_updates,
                    resources.inventory_source_updates,
                    resources.project_update_scm_inventory_updates],
                   InventoryUpdates)


class InventoryUpdateCancel(base.Base):

    pass


page.register_page(resources.inventory_update_cancel, InventoryUpdateCancel)


class InventoryCopy(base.Base):

    pass


page.register_page(resources.inventory_copy, InventoryCopy)



from awxkit.api.resources import resources
from . import users
from . import page


class AccessList(page.PageList, users.User):

    pass


page.register_page([resources.organization_access_list,
                    resources.user_access_list,
                    resources.inventory_access_list,
                    resources.group_access_list,
                    resources.credential_access_list,
                    resources.project_access_list,
                    resources.job_template_access_list,
                    resources.team_access_list], AccessList)



import threading
import logging
import atexit
import json
import ssl

from queue import Queue, Empty
from urllib.parse import urlparse

from awxkit.config import config


log = logging.getLogger(__name__)


class WSClientException(Exception):

    pass


changed = 'changed'
limit_reached = 'limit_reached'
status_changed = 'status_changed'
summary = 'summary'


class WSClient(object):
    """Provides a basic means of testing pub/sub notifications with payloads similar to
    'groups': {'jobs': ['status_changed', 'summary'],
               'schedules': ['changed'],
               'ad_hoc_command_events': [ids...],
               'job_events': [ids...],
               'workflow_events': [ids...],
               'project_update_events': [ids...],
               'inventory_update_events': [ids...],
               'system_job_events': [ids...],
               'control': ['limit_reached']}
    e.x:
    ```
    ws = WSClient(token, port=8013, secure=False).connect()
    ws.job_details()
    ... # launch job
    job_messages = [msg for msg in ws]
    ws.ad_hoc_stdout()
    ... # launch ad hoc command
    ad_hoc_messages = [msg for msg in ws]
    ws.close()
    ```
    """

    # Subscription group types

    def __init__(self, token=None, hostname='', port=443, secure=True, session_id=None, csrftoken=None):
        # delay this import, because this is an optional dependency
        import websocket

        if not hostname:
            result = urlparse(config.base_url)
            secure = result.scheme == 'https'
            port = result.port
            if port is None:
                port = 80
                if secure:
                    port = 443
            # should we be adding result.path here?
            hostname = result.hostname

        self.port = port
        self._use_ssl = secure
        self.hostname = hostname
        self.token = token
        self.session_id = session_id
        self.csrftoken = csrftoken
        self._recv_queue = Queue()
        self._ws_closed = False
        self._ws_connected_flag = threading.Event()
        if self.token is not None:
            auth_cookie = 'token="{0.token}";'.format(self)
        elif self.session_id is not None:
            auth_cookie = 'sessionid="{0.session_id}"'.format(self)
            if self.csrftoken:
                auth_cookie += ';csrftoken={0.csrftoken}'.format(self)
        else:
            auth_cookie = ''
        pref = 'wss://' if self._use_ssl else 'ws://'
        url = '{0}{1.hostname}:{1.port}/websocket/'.format(pref, self)
        self.ws = websocket.WebSocketApp(url,
                                         on_open=self._on_open,
                                         on_message=self._on_message,
                                         on_error=self._on_error,
                                         on_close=self._on_close,
                                         cookie=auth_cookie)
        self._message_cache = []
        self._should_subscribe_to_pending_job = False
        self._pending_unsubscribe = threading.Event()

    def connect(self):
        wst = threading.Thread(target=self._ws_run_forever, args=(self.ws, {"cert_reqs": ssl.CERT_NONE}))
        wst.daemon = True
        wst.start()
        atexit.register(self.close)
        if not self._ws_connected_flag.wait(20):
            raise WSClientException('Failed to establish channel connection w/ AWX.')
        return self

    def close(self):
        log.info('close method was called, but ignoring')
        if not self._ws_closed:
            log.info('Closing websocket connection.')
            self.ws.close()

    def job_details(self, *job_ids):
        """subscribes to job status, summary, and, for the specified ids, job events"""
        self.subscribe(jobs=[status_changed, summary], job_events=list(job_ids))

    def pending_job_details(self):
        """subscribes to job status and summary, with responsive
        job event subscription for an id provided by AWX
        """
        self.subscribe_to_pending_events('job_events', [status_changed, summary])

    def status_changes(self):
        self.subscribe(jobs=[status_changed])

    def job_stdout(self, *job_ids):
        self.subscribe(jobs=[status_changed], job_events=list(job_ids))

    def pending_job_stdout(self):
        self.subscribe_to_pending_events('job_events')

    # mirror page behavior
    def ad_hoc_stdout(self, *ahc_ids):
        self.subscribe(jobs=[status_changed], ad_hoc_command_events=list(ahc_ids))

    def pending_ad_hoc_stdout(self):
        self.subscribe_to_pending_events('ad_hoc_command_events')

    def project_update_stdout(self, *project_update_ids):
        self.subscribe(jobs=[status_changed], project_update_events=list(project_update_ids))

    def pending_project_update_stdout(self):
        self.subscribe_to_pending_events('project_update_events')

    def inventory_update_stdout(self, *inventory_update_ids):
        self.subscribe(jobs=[status_changed], inventory_update_events=list(inventory_update_ids))

    def pending_inventory_update_stdout(self):
        self.subscribe_to_pending_events('inventory_update_events')

    def workflow_events(self, *wfjt_ids):
        self.subscribe(jobs=[status_changed], workflow_events=list(wfjt_ids))

    def pending_workflow_events(self):
        self.subscribe_to_pending_events('workflow_events')

    def system_job_events(self, *system_job_ids):
        self.subscribe(jobs=[status_changed], system_job_events=list(system_job_ids))

    def pending_system_job_events(self):
        self.subscribe_to_pending_events('system_job_events')

    def subscribe_to_pending_events(self, events, jobs=[status_changed]):
        self._should_subscribe_to_pending_job = dict(jobs=jobs, events=events)
        self.subscribe(jobs=jobs)

    # mirror page behavior
    def jobs_list(self):
        self.subscribe(jobs=[status_changed, summary], schedules=[changed])

    # mirror page behavior
    def dashboard(self):
        self.subscribe(jobs=[status_changed])

    def subscribe(self, **groups):
        """Sends a subscription request for the specified channel groups.
        ```
        ws.subscribe(jobs=[ws.status_changed, ws.summary],
                     job_events=[1,2,3])
        ```
        """
        self._subscribe(groups=groups)

    def _subscribe(self, **payload):
        payload['xrftoken'] = self.csrftoken
        self._send(json.dumps(payload))

    def unsubscribe(self, wait=True, timeout=10):
        if wait:
            # Other unnsubscribe events could have caused the edge to trigger.
            # This way the _next_ event will trigger our waiting.
            self._pending_unsubscribe.clear()
            self._send(json.dumps(dict(groups={}, xrftoken=self.csrftoken)))
            if not self._pending_unsubscribe.wait(timeout):
                raise RuntimeError("Failed while waiting on unsubscribe reply because timeout of {} seconds was reached.".format(timeout))
        else:
            self._send(json.dumps(dict(groups={}, xrftoken=self.csrftoken)))

    def _on_message(self, message):
        message = json.loads(message)
        log.debug('received message: {}'.format(message))

        if all([message.get('group_name') == 'jobs',
                message.get('status') == 'pending',
                message.get('unified_job_id'),
                self._should_subscribe_to_pending_job]):
            if bool(message.get('project_id')) == (
                    self._should_subscribe_to_pending_job['events'] == 'project_update_events'):
                self._update_subscription(message['unified_job_id'])

        ret = self._recv_queue.put(message)

        # unsubscribe acknowledgement
        if 'groups_current' in message:
            self._pending_unsubscribe.set()

        return ret

    def _update_subscription(self, job_id):
        subscription = dict(jobs=self._should_subscribe_to_pending_job['jobs'])
        events = self._should_subscribe_to_pending_job['events']
        subscription[events] = [job_id]
        self.subscribe(**subscription)
        self._should_subscribe_to_pending_job = False

    def _on_open(self):
        self._ws_connected_flag.set()

    def _on_error(self, error):
        log.info('Error received: {}'.format(error))

    def _on_close(self):
        log.info('Successfully closed ws.')
        self._ws_closed = True

    def _ws_run_forever(self, sockopt=None, sslopt=None):
        self.ws.run_forever(sslopt=sslopt)
        log.debug('ws.run_forever finished')

    def _recv(self, wait=False, timeout=10):
        try:
            msg = self._recv_queue.get(wait, timeout)
        except Empty:
            return None
        return msg

    def _send(self, data):
        self.ws.send(data)
        log.debug('successfully sent {}'.format(data))

    def __iter__(self):
        while True:
            val = self._recv()
            if not val:
                return
            yield val



import argparse
import json
import unittest
from io import StringIO

import pytest
from requests import Response

from awxkit.api.pages import Page
from awxkit.cli.options import ResourceOptionsParser


class ResourceOptionsParser(ResourceOptionsParser):

    def get_allowed_options(self):
        self.allowed_options = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']


class OptionsPage(Page):

    def options(self):
        return self

    def endswith(self, v):
        return self.endpoint.endswith(v)

    def __getitem__(self, k):
        return {
            'GET': {},
            'POST': {},
            'PUT': {},
        }


class TestOptions(unittest.TestCase):

    def setUp(self):
        _parser = argparse.ArgumentParser()
        self.parser = _parser.add_subparsers(help='action')

    def test_list(self):
        page = OptionsPage.from_json({
            'actions': {
                'GET': {},
                'POST': {},
            }
        })
        ResourceOptionsParser(None, page, 'users', self.parser)
        assert 'list' in self.parser.choices

    def test_list_filtering(self):
        page = OptionsPage.from_json({
            'actions': {
                'GET': {},
                'POST': {
                    'first_name': {'type': 'string'}
                },
            }
        })
        options = ResourceOptionsParser(None, page, 'users', self.parser)
        options.build_query_arguments('list', 'POST')
        assert 'list' in self.parser.choices

        out = StringIO()
        self.parser.choices['list'].print_help(out)
        assert '--first_name TEXT' in out.getvalue()

    def test_list_not_filterable(self):
        page = OptionsPage.from_json({
            'actions': {
                'GET': {},
                'POST': {
                    'middle_name': {'type': 'string', 'filterable': False}
                },
            }
        })
        options = ResourceOptionsParser(None, page, 'users', self.parser)
        options.build_query_arguments('list', 'POST')
        assert 'list' in self.parser.choices

        out = StringIO()
        self.parser.choices['list'].print_help(out)
        assert '--middle_name' not in out.getvalue()

    def test_creation_optional_argument(self):
        page = OptionsPage.from_json({
            'actions': {
                'POST': {
                    'first_name': {
                        'type': 'string',
                        'help_text': 'Please specify your first name',
                    }
                },
            }
        })
        options = ResourceOptionsParser(None, page, 'users', self.parser)
        options.build_query_arguments('create', 'POST')
        assert 'create' in self.parser.choices

        out = StringIO()
        self.parser.choices['create'].print_help(out)
        assert '--first_name TEXT  Please specify your first name' in out.getvalue()

    def test_creation_required_argument(self):
        page = OptionsPage.from_json({
            'actions': {
                'POST': {
                    'username': {
                        'type': 'string',
                        'help_text': 'Please specify a username',
                        'required': True
                    }
                },
            }
        })
        options = ResourceOptionsParser(None, page, 'users', self.parser)
        options.build_query_arguments('create', 'POST')
        assert 'create' in self.parser.choices

        out = StringIO()
        self.parser.choices['create'].print_help(out)
        assert '--username TEXT  Please specify a username'

    def test_integer_argument(self):
        page = OptionsPage.from_json({
            'actions': {
                'POST': {
                    'max_hosts': {'type': 'integer'}
                },
            }
        })
        options = ResourceOptionsParser(None, page, 'organizations', self.parser)
        options.build_query_arguments('create', 'POST')
        assert 'create' in self.parser.choices

        out = StringIO()
        self.parser.choices['create'].print_help(out)
        assert '--max_hosts INTEGER' in out.getvalue()

    def test_boolean_argument(self):
        page = OptionsPage.from_json({
            'actions': {
                'POST': {
                    'diff_mode': {'type': 'boolean'}
                },
            }
        })
        options = ResourceOptionsParser(None, page, 'users', self.parser)
        options.build_query_arguments('create', 'POST')
        assert 'create' in self.parser.choices

        out = StringIO()
        self.parser.choices['create'].print_help(out)
        assert '--diff_mode BOOLEAN' in out.getvalue()

    def test_choices(self):
        page = OptionsPage.from_json({
            'actions': {
                'POST': {
                    'verbosity': {
                        'type': 'integer',
                        'choices': [
                            (0, '0 (Normal)'),
                            (1, '1 (Verbose)'),
                            (2, '2 (More Verbose)'),
                            (3, '3 (Debug)'),
                            (4, '4 (Connection Debug)'),
                            (5, '5 (WinRM Debug)'),
                        ]
                    }
                },
            }
        })
        options = ResourceOptionsParser(None, page, 'users', self.parser)
        options.build_query_arguments('create', 'POST')
        assert 'create' in self.parser.choices

        out = StringIO()
        self.parser.choices['create'].print_help(out)
        assert '--verbosity {0,1,2,3,4,5}' in out.getvalue()

    def test_actions_with_primary_key(self):
        for method in ('get', 'modify', 'delete'):
            page = OptionsPage.from_json({
                'actions': {'GET': {}, 'POST': {}}
            })
            ResourceOptionsParser(None, page, 'jobs', self.parser)
            assert method in self.parser.choices

            out = StringIO()
            self.parser.choices[method].print_help(out)
            assert 'positional arguments:\n  id' in out.getvalue()


class TestSettingsOptions(unittest.TestCase):

    def setUp(self):
        _parser = argparse.ArgumentParser()
        self.parser = _parser.add_subparsers(help='action')

    def test_list(self):
        page = OptionsPage.from_json({
            'actions': {
                'GET': {},
                'POST': {},
                'PUT': {},
            }
        })
        page.endpoint = '/settings/all/'
        ResourceOptionsParser(None, page, 'settings', self.parser)
        assert 'list' in self.parser.choices
        assert 'modify' in self.parser.choices

        out = StringIO()
        self.parser.choices['modify'].print_help(out)
        assert 'modify [-h] key value' in out.getvalue()



# list of random English nouns used for resource name utilities

words = ['People', 'History', 'Way', 'Art', 'World', 'Information', 'Map', 'Two',
         'Family', 'Government', 'Health', 'System', 'Computer', 'Meat', 'Year', 'Thanks',
         'Music', 'Person', 'Reading', 'Method', 'Data', 'Food', 'Understanding', 'Theory',
         'Law', 'Bird', 'Literature', 'Problem', 'Software', 'Control', 'Knowledge', 'Power',
         'Ability', 'Economics', 'Love', 'Internet', 'Television', 'Science', 'Library', 'Nature',
         'Fact', 'Product', 'Idea', 'Temperature', 'Investment', 'Area', 'Society', 'Activity',
         'Story', 'Industry', 'Media', 'Thing', 'Oven', 'Community', 'Definition', 'Safety',
         'Quality', 'Development', 'Language', 'Management', 'Player', 'Variety', 'Video', 'Week',
         'Security', 'Country', 'Exam', 'Movie', 'Organization', 'Equipment', 'Physics', 'Analysis',
         'Policy', 'Series', 'Thought', 'Basis', 'Boyfriend', 'Direction', 'Strategy', 'Technology',
         'Army', 'Camera', 'Freedom', 'Paper', 'Environment', 'Child', 'Instance', 'Month',
         'Truth', 'Marketing', 'University', 'Writing', 'Article', 'Department', 'Difference', 'Goal',
         'News', 'Audience', 'Fishing', 'Growth', 'Income', 'Marriage', 'User', 'Combination',
         'Failure', 'Meaning', 'Medicine', 'Philosophy', 'Teacher', 'Communication', 'Night', 'Chemistry',
         'Disease', 'Disk', 'Energy', 'Nation', 'Road', 'Role', 'Soup', 'Advertising',
         'Location', 'Success', 'Addition', 'Apartment', 'Education', 'Math', 'Moment', 'Painting',
         'Politics', 'Attention', 'Decision', 'Event', 'Property', 'Shopping', 'Student', 'Wood',
         'Competition', 'Distribution', 'Entertainment', 'Office', 'Population', 'President', 'Unit', 'Category',
         'Cigarette', 'Context', 'Introduction', 'Opportunity', 'Performance', 'Driver', 'Flight', 'Length',
         'Magazine', 'Newspaper', 'Relationship', 'Teaching', 'Cell', 'Dealer', 'Debate', 'Finding',
         'Lake', 'Member', 'Message', 'Phone', 'Scene', 'Appearance', 'Association', 'Concept',
         'Customer', 'Death', 'Discussion', 'Housing', 'Inflation', 'Insurance', 'Mood', 'Woman',
         'Advice', 'Blood', 'Effort', 'Expression', 'Importance', 'Opinion', 'Payment', 'Reality',
         'Responsibility', 'Situation', 'Skill', 'Statement', 'Wealth', 'Application', 'City', 'County',
         'Depth', 'Estate', 'Foundation', 'Grandmother', 'Heart', 'Perspective', 'Photo', 'Recipe',
         'Studio', 'Topic', 'Collection', 'Depression', 'Imagination', 'Passion', 'Percentage', 'Resource',
         'Setting', 'Ad', 'Agency', 'College', 'Connection', 'Criticism', 'Debt', 'Description',
         'Memory', 'Patience', 'Secretary', 'Solution', 'Administration', 'Aspect', 'Attitude', 'Director',
         'Personality', 'Psychology', 'Recommendation', 'Response', 'Selection', 'Storage', 'Version', 'Alcohol',
         'Argument', 'Complaint', 'Contract', 'Emphasis', 'Highway', 'Loss', 'Membership', 'Possession',
         'Preparation', 'Steak', 'Union', 'Agreement', 'Cancer', 'Currency', 'Employment', 'Engineering',
         'Entry', 'Interaction', 'Limit', 'Mixture', 'Preference', 'Region', 'Republic', 'Seat',
         'Tradition', 'Virus', 'Actor', 'Classroom', 'Delivery', 'Device', 'Difficulty', 'Drama',
         'Election', 'Engine', 'Football', 'Guidance', 'Hotel', 'Match', 'Owner', 'Priority',
         'Protection', 'Suggestion', 'Tension', 'Variation', 'Anxiety', 'Atmosphere', 'Awareness', 'Bread',
         'Climate', 'Comparison', 'Confusion', 'Construction', 'Elevator', 'Emotion', 'Employee', 'Employer',
         'Guest', 'Height', 'Leadership', 'Mall', 'Manager', 'Operation', 'Recording', 'Respect',
         'Sample', 'Transportation', 'Boring', 'Charity', 'Cousin', 'Disaster', 'Editor', 'Efficiency',
         'Excitement', 'Extent', 'Feedback', 'Guitar', 'Homework', 'Leader', 'Mom', 'Outcome',
         'Permission', 'Presentation', 'Promotion', 'Reflection', 'Refrigerator', 'Resolution', 'Revenue', 'Session',
         'Singer', 'Tennis', 'Basket', 'Bonus', 'Cabinet', 'Childhood', 'Church', 'Clothes',
         'Coffee', 'Dinner', 'Drawing', 'Hair', 'Hearing', 'Initiative', 'Judgment', 'Lab',
         'Measurement', 'Mode', 'Mud', 'Orange', 'Poetry', 'Police', 'Possibility', 'Procedure',
         'Queen', 'Ratio', 'Relation', 'Restaurant', 'Satisfaction', 'Sector', 'Signature', 'Significance',
         'Song', 'Tooth', 'Town', 'Vehicle', 'Volume', 'Wife', 'Accident', 'Airport',
         'Appointment', 'Arrival', 'Assumption', 'Baseball', 'Chapter', 'Committee', 'Conversation', 'Database',
         'Enthusiasm', 'Error', 'Explanation', 'Farmer', 'Gate', 'Girl', 'Hall', 'Historian',
         'Hospital', 'Injury', 'Instruction', 'Maintenance', 'Manufacturer', 'Meal', 'Perception', 'Pie',
         'Poem', 'Presence', 'Proposal', 'Reception', 'Replacement', 'Revolution', 'River', 'Son',
         'Speech', 'Tea', 'Village', 'Warning', 'Winner', 'Worker', 'Writer', 'Assistance',
         'Breath', 'Buyer', 'Chest', 'Chocolate', 'Conclusion', 'Contribution', 'Cookie', 'Courage',
         'Dad', 'Desk', 'Drawer', 'Establishment', 'Examination', 'Garbage', 'Grocery', 'Honey',
         'Impression', 'Improvement', 'Independence', 'Insect', 'Inspection', 'Inspector', 'King', 'Ladder',
         'Menu', 'Penalty', 'Piano', 'Potato', 'Profession', 'Professor', 'Quantity', 'Reaction',
         'Requirement', 'Salad', 'Sister', 'Supermarket', 'Tongue', 'Weakness', 'Wedding', 'Affair',
         'Ambition', 'Analyst', 'Apple', 'Assignment', 'Assistant', 'Bathroom', 'Bedroom', 'Beer',
         'Birthday', 'Celebration', 'Championship', 'Cheek', 'Client', 'Consequence', 'Departure', 'Diamond',
         'Dirt', 'Ear', 'Fortune', 'Friendship', 'Snapewife', 'Funeral', 'Gene', 'Girlfriend', 'Hat',
         'Indication', 'Intention', 'Lady', 'Midnight', 'Negotiation', 'Obligation', 'Passenger', 'Pizza',
         'Platform', 'Poet', 'Pollution', 'Recognition', 'Reputation', 'Shirt', 'Sir', 'Speaker',
         'Stranger', 'Surgery', 'Sympathy', 'Tale', 'Throat', 'Trainer', 'Uncle', 'Youth',
         'Time', 'Work', 'Film', 'Water', 'Money', 'Example', 'While', 'Business',
         'Study', 'Game', 'Life', 'Form', 'Air', 'Day', 'Place', 'Number',
         'Part', 'Field', 'Fish', 'Back', 'Process', 'Heat', 'Hand', 'Experience',
         'Job', 'Book', 'End', 'Point', 'Type', 'Home', 'Economy', 'Value',
         'Body', 'Market', 'Guide', 'Interest', 'State', 'Radio', 'Course', 'Company',
         'Price', 'Size', 'Card', 'List', 'Mind', 'Trade', 'Line', 'Care',
         'Group', 'Risk', 'Word', 'Fat', 'Force', 'Key', 'Light', 'Training',
         'Name', 'School', 'Top', 'Amount', 'Level', 'Order', 'Practice', 'Research',
         'Sense', 'Service', 'Piece', 'Web', 'Boss', 'Sport', 'Fun', 'House',
         'Page', 'Term', 'Test', 'Answer', 'Sound', 'Focus', 'Matter', 'Kind',
         'Soil', 'Board', 'Oil', 'Picture', 'Access', 'Garden', 'Range', 'Rate',
         'Reason', 'Future', 'Site', 'Demand', 'Exercise', 'Image', 'Case', 'Cause',
         'Coast', 'Action', 'Age', 'Bad', 'Boat', 'Record', 'Result', 'Section',
         'Building', 'Mouse', 'Cash', 'Class', 'Nothing', 'Period', 'Plan', 'Store',
         'Tax', 'Side', 'Subject', 'Space', 'Rule', 'Stock', 'Weather', 'Chance',
         'Figure', 'Man', 'Model', 'Source', 'Beginning', 'Earth', 'Program', 'Chicken',
         'Design', 'Feature', 'Head', 'Material', 'Purpose', 'Question', 'Rock', 'Salt',
         'Act', 'Birth', 'Car', 'Dog', 'Object', 'Scale', 'Sun', 'Note',
         'Profit', 'Rent', 'Speed', 'Style', 'War', 'Bank', 'Craft', 'Half',
         'Inside', 'Outside', 'Standard', 'Bus', 'Exchange', 'Eye', 'Fire', 'Position',
         'Pressure', 'Stress', 'Advantage', 'Benefit', 'Box', 'Frame', 'Issue', 'Step',
         'Cycle', 'Face', 'Item', 'Metal', 'Paint', 'Review', 'Room', 'Screen',
         'Structure', 'View', 'Account', 'Ball', 'Discipline', 'Medium', 'Share', 'Balance',
         'Bit', 'Black', 'Bottom', 'Choice', 'Gift', 'Impact', 'Machine', 'Shape',
         'Tool', 'Wind', 'Address', 'Average', 'Career', 'Culture', 'Morning', 'Pot',
         'Sign', 'Table', 'Task', 'Condition', 'Contact', 'Credit', 'Egg', 'Hope',
         'Ice', 'Network', 'North', 'Square', 'Attempt', 'Date', 'Effect', 'Link',
         'Post', 'Star', 'Voice', 'Capital', 'Challenge', 'Friend', 'Self', 'Shot',
         'Brush', 'Couple', 'Exit', 'Front', 'Function', 'Lack', 'Living', 'Plant',
         'Plastic', 'Spot', 'Summer', 'Taste', 'Theme', 'Track', 'Wing', 'Brain',
         'Button', 'Click', 'Desire', 'Foot', 'Gas', 'Influence', 'Notice', 'Rain',
         'Wall', 'Base', 'Damage', 'Distance', 'Feeling', 'Pair', 'Savings', 'Staff',
         'Sugar', 'Target', 'Text', 'Animal', 'Author', 'Budget', 'Discount', 'File',
         'Ground', 'Lesson', 'Minute', 'Officer', 'Phase', 'Reference', 'Register', 'Sky',
         'Stage', 'Stick', 'Title', 'Trouble', 'Bowl', 'Bridge', 'Campaign', 'Character',
         'Club', 'Edge', 'Evidence', 'Fan', 'Letter', 'Lock', 'Maximum', 'Novel',
         'Option', 'Pack', 'Park', 'Plenty', 'Quarter', 'Skin', 'Sort', 'Weight',
         'Baby', 'Background', 'Carry', 'Dish', 'Factor', 'Fruit', 'Glass', 'Joint',
         'Master', 'Muscle', 'Red', 'Strength', 'Traffic', 'Trip', 'Vegetable', 'Appeal',
         'Chart', 'Gear', 'Ideal', 'Kitchen', 'Land', 'Log', 'Mother', 'Net',
         'Party', 'Principle', 'Relative', 'Sale', 'Season', 'Signal', 'Spirit', 'Street',
         'Tree', 'Wave', 'Belt', 'Bench', 'Commission', 'Copy', 'Drop', 'Minimum',
         'Path', 'Progress', 'Project', 'Sea', 'South', 'Status', 'Stuff', 'Ticket',
         'Tour', 'Angle', 'Blue', 'Breakfast', 'Confidence', 'Daughter', 'Degree', 'Doctor',
         'Dot', 'Dream', 'Duty', 'Essay', 'Father', 'Fee', 'Finance', 'Hour',
         'Juice', 'Luck', 'Milk', 'Mouth', 'Peace', 'Pipe', 'Stable', 'Storm',
         'Substance', 'Team', 'Trick', 'Afternoon', 'Bat', 'Beach', 'Blank', 'Catch',
         'Chain', 'Consideration', 'Cream', 'Crew', 'Detail', 'Gold', 'Interview', 'Kid',
         'Mark', 'Mission', 'Pain', 'Pleasure', 'Score', 'Screw', 'Gratitude', 'Shop',
         'Shower', 'Suit', 'Tone', 'Window', 'Agent', 'Band', 'Bath', 'Block',
         'Bone', 'Calendar', 'Candidate', 'Cap', 'Coat', 'Contest', 'Corner', 'Court',
         'Cup', 'District', 'Door', 'East', 'Finger', 'Garage', 'Guarantee', 'Hole',
         'Hook', 'Implement', 'Layer', 'Lecture', 'Lie', 'Manner', 'Meeting', 'Nose',
         'Parking', 'Partner', 'Profile', 'Rice', 'Routine', 'Schedule', 'Swimming', 'Telephone',
         'Tip', 'Winter', 'Airline', 'Bag', 'Battle', 'Bed', 'Bill', 'Bother',
         'Cake', 'Code', 'Curve', 'Designer', 'Dimension', 'Dress', 'Ease', 'Emergency',
         'Evening', 'Extension', 'Farm', 'Fight', 'Gap', 'Grade', 'Holiday', 'Horror',
         'Horse', 'Host', 'Husband', 'Loan', 'Mistake', 'Mountain', 'Nail', 'Noise',
         'Occasion', 'Package', 'Patient', 'Pause', 'Phrase', 'Proof', 'Race', 'Relief',
         'Sand', 'Sentence', 'Shoulder', 'Smoke', 'Stomach', 'String', 'Tourist', 'Towel',
         'Vacation', 'West', 'Wheel', 'Wine', 'Arm', 'Aside', 'Associate', 'Bet',
         'Blow', 'Border', 'Branch', 'Breast', 'Brother', 'Buddy', 'Bunch', 'Chip',
         'Coach', 'Cross', 'Document', 'Draft', 'Dust', 'Expert', 'Floor', 'God',
         'Golf', 'Habit', 'Iron', 'Judge', 'Knife', 'Landscape', 'League', 'Mail',
         'Mess', 'Native', 'Opening', 'Parent', 'Pattern', 'Pin', 'Pool', 'Pound',
         'Request', 'Salary', 'Shame', 'Shelter', 'Shoe', 'Silver', 'Tackle', 'Tank',
         'Trust', 'Assist', 'Bake', 'Bar', 'Bell', 'Bike', 'Blame', 'Boy',
         'Brick', 'Chair', 'Closet', 'Clue', 'Collar', 'Comment', 'Conference', 'Devil',
         'Diet', 'Fear', 'Fuel', 'Glove', 'Jacket', 'Lunch', 'Monitor', 'Mortgage',
         'Nurse', 'Pace', 'Panic', 'Peak', 'Plane', 'Reward', 'Row', 'Sandwich',
         'Shock', 'Spite', 'Spray', 'Surprise', 'Till', 'Transition', 'Weekend', 'Welcome',
         'Yard', 'Alarm', 'Bend', 'Bicycle', 'Bite', 'Blind', 'Bottle', 'Cable',
         'Candle', 'Clerk', 'Cloud', 'Concert', 'Counter', 'Flower', 'Grandfather', 'Harm',
         'Knee', 'Lawyer', 'Leather', 'Load', 'Mirror', 'Neck', 'Pension', 'Plate',
         'Purple', 'Ruin', 'Ship', 'Skirt', 'Slice', 'Snow', 'Specialist', 'Stroke',
         'Switch', 'Trash', 'Tune', 'Zone', 'Anger', 'Award', 'Bid', 'Bitter',
         'Boot', 'Bug', 'Camp', 'Candy', 'Carpet', 'Cat', 'Champion', 'Channel',
         'Clock', 'Comfort', 'Cow', 'Crack', 'Engineer', 'Entrance', 'Fault', 'Grass',
         'Guy', 'Hell', 'Highlight', 'Incident', 'Island', 'Joke', 'Jury', 'Leg',
         'Lip', 'Mate', 'Motor', 'Nerve', 'Passage', 'Pen', 'Pride', 'Priest',
         'Prize', 'Promise', 'Resident', 'Resort', 'Ring', 'Roof', 'Rope', 'Sail',
         'Scheme', 'Script', 'Sock', 'Station', 'Toe', 'Tower', 'Truck', 'Witness',
         'Asparagus', 'You', 'It', 'Can', 'Will', 'If', 'One', 'Many',
         'Most', 'Other', 'Use', 'Make', 'Good', 'Look', 'Help', 'Go',
         'Great', 'Being', 'Few', 'Might', 'Still', 'Public', 'Read', 'Keep',
         'Start', 'Give', 'Human', 'Local', 'General', 'She', 'Specific', 'Long',
         'Play', 'Feel', 'High', 'Tonight', 'Put', 'Common', 'Set', 'Change',
         'Simple', 'Past', 'Big', 'Possible', 'Particular', 'Today', 'Major', 'Personal',
         'Current', 'National', 'Cut', 'Natural', 'Physical', 'Show', 'Try', 'Check',
         'Second', 'Call', 'Move', 'Pay', 'Let', 'Increase', 'Single', 'Individual',
         'Turn', 'Ask', 'Buy', 'Guard', 'Hold', 'Main', 'Offer', 'Potential',
         'Professional', 'International', 'Travel', 'Cook', 'Alternative', 'Following', 'Special', 'Working',
         'Whole', 'Dance', 'Excuse', 'Cold', 'Commercial', 'Low', 'Purchase', 'Deal',
         'Primary', 'Worth', 'Fall', 'Necessary', 'Positive', 'Produce', 'Search', 'Present',
         'Spend', 'Talk', 'Creative', 'Tell', 'Cost', 'Drive', 'Green', 'Support',
         'Glad', 'Remove', 'Return', 'Run', 'Complex', 'Due', 'Effective', 'Middle',
         'Regular', 'Reserve', 'Independent', 'Leave', 'Original', 'Reach', 'Rest', 'Serve',
         'Watch', 'Beautiful', 'Charge', 'Active', 'Break', 'Negative', 'Safe', 'Stay',
         'Visit', 'Visual', 'Affect', 'Cover', 'Report', 'Rise', 'Walk', 'White',
         'Beyond', 'Junior', 'Pick', 'Unique', 'Anything', 'Classic', 'Final', 'Lift',
         'Mix', 'Private', 'Stop', 'Teach', 'Western', 'Concern', 'Familiar', 'Fly',
         'Official', 'Broad', 'Comfortable', 'Gain', 'Maybe', 'Rich', 'Save', 'Stand',
         'Young', 'Heavy', 'Hello', 'Lead', 'Listen', 'Valuable', 'Worry', 'Handle',
         'Leading', 'Meet', 'Release', 'Sell', 'Finish', 'Normal', 'Press', 'Ride',
         'Secret', 'Spread', 'Spring', 'Tough', 'Wait', 'Brown', 'Deep', 'Display',
         'Flow', 'Hit', 'Objective', 'Shoot', 'Touch', 'Cancel', 'Chemical', 'Cry',
         'Dump', 'Extreme', 'Push', 'Conflict', 'Eat', 'Fill', 'Formal', 'Jump',
         'Kick', 'Opposite', 'Pass', 'Pitch', 'Remote', 'Total', 'Treat', 'Vast',
         'Abuse', 'Beat', 'Burn', 'Deposit', 'Print', 'Raise', 'Sleep', 'Somewhere',
         'Advance', 'Anywhere', 'Consist', 'Dark', 'Double', 'Draw', 'Equal', 'Fix',
         'Hire', 'Internal', 'Join', 'Kill', 'Sensitive', 'Tap', 'Win', 'Attack',
         'Claim', 'Constant', 'Drag', 'Drink', 'Guess', 'Minor', 'Pull', 'Raw',
         'Soft', 'Solid', 'Wear', 'Weird', 'Wonder', 'Annual', 'Count', 'Dead',
         'Doubt', 'Feed', 'Forever', 'Impress', 'Nobody', 'Repeat', 'Round', 'Sing',
         'Slide', 'Strip', 'Whereas', 'Wish', 'Combine', 'Command', 'Dig', 'Divide',
         'Equivalent', 'Hang', 'Hunt', 'Initial', 'March', 'Mention', 'Spiritual', 'Survey',
         'Tie', 'Adult', 'Brief', 'Crazy', 'Escape', 'Gather', 'Hate', 'Prior',
         'Repair', 'Rough', 'Sad', 'Scratch', 'Sick', 'Strike', 'Employ', 'External',
         'Hurt', 'Illegal', 'Laugh', 'Lay', 'Mobile', 'Nasty', 'Ordinary', 'Respond',
         'Royal', 'Senior', 'Split', 'Strain', 'Struggle', 'Swim', 'Train', 'Upper',
         'Wash', 'Yellow', 'Convert', 'Crash', 'Dependent', 'Fold', 'Funny', 'Grab',
         'Hide', 'Miss', 'Permit', 'Quote', 'Recover', 'Resolve', 'Roll', 'Sink',
         'Slip', 'Spare', 'Suspect', 'Sweet', 'Swing', 'Twist', 'Upstairs', 'Usual',
         'Abroad', 'Brave', 'Calm', 'Concentrate', 'Estimate', 'Grand', 'Male', 'Mine',
         'Prompt', 'Quiet', 'Refuse', 'Regret', 'Reveal', 'Rush', 'Shake', 'Shift',
         'Shine', 'Steal', 'Suck', 'Surround', 'Anybody', 'Bear', 'Brilliant', 'Dare',
         'Dear', 'Delay', 'Drunk', 'Female', 'Hurry', 'Inevitable', 'Invite', 'Kiss',
         'Neat', 'Pop', 'Punch', 'Quit', 'Reply', 'Representative', 'Resist', 'Rip',
         'Rub', 'Silly', 'Smile', 'Spell', 'Stretch', 'Stupid', 'Tear', 'Temporary',
         'Tomorrow', 'Wake', 'Wrap', 'Yesterday']



import yaml
import json
import os

from awxkit import api, config
from awxkit.exceptions import ImportExportError
from awxkit.utils import to_str
from awxkit.api.pages import Page
from awxkit.api.pages.api import EXPORTABLE_RESOURCES
from awxkit.cli.format import FORMATTERS, format_response, add_authentication_arguments
from awxkit.cli.utils import CustomRegistryMeta, cprint


CONTROL_RESOURCES = ['ping', 'config', 'me', 'metrics']

DEPRECATED_RESOURCES = {
    'ad_hoc_commands': 'ad_hoc',
    'applications': 'application',
    'credentials': 'credential',
    'credential_types': 'credential_type',
    'groups': 'group',
    'hosts': 'host',
    'instances': 'instance',
    'instance_groups': 'instance_group',
    'inventory': 'inventories',
    'inventory_scripts': 'inventory_script',
    'inventory_sources': 'inventory_source',
    'inventory_updates': 'inventory_update',
    'jobs': 'job',
    'job_templates': 'job_template',
    'labels': 'label',
    'workflow_job_template_nodes': 'node',
    'notification_templates': 'notification_template',
    'organizations': 'organization',
    'projects': 'project',
    'project_updates': 'project_update',
    'roles': 'role',
    'schedules': 'schedule',
    'settings': 'setting',
    'teams': 'team',
    'workflow_job_templates': 'workflow',
    'workflow_jobs': 'workflow_job',
    'users': 'user'
}
DEPRECATED_RESOURCES_REVERSE = dict(
    (v, k) for k, v in DEPRECATED_RESOURCES.items()
)


class CustomCommand(metaclass=CustomRegistryMeta):
    """Base class for implementing custom commands.

    Custom commands represent static code which should run - they are
    responsible for returning and formatting their own output (which may or may
    not be JSON/YAML).
    """

    help_text = ''

    @property
    def name(self):
        raise NotImplementedError()

    def handle(self, client, parser):
        """To be implemented by subclasses.
        Should return a dictionary that is JSON serializable
        """
        raise NotImplementedError()


class Login(CustomCommand):
    name = 'login'
    help_text = 'authenticate and retrieve an OAuth2 token'

    def print_help(self, parser):
        add_authentication_arguments(parser, os.environ)
        parser.print_help()

    def handle(self, client, parser):
        auth = parser.add_argument_group('OAuth2.0 Options')
        auth.add_argument('--description', help='description of the generated OAuth2.0 token', metavar='TEXT')
        auth.add_argument('--conf.client_id', metavar='TEXT')
        auth.add_argument('--conf.client_secret', metavar='TEXT')
        auth.add_argument(
            '--conf.scope', choices=['read', 'write'], default='write'
        )
        if client.help:
            self.print_help(parser)
            raise SystemExit()
        parsed = parser.parse_known_args()[0]
        kwargs = {
            'client_id': getattr(parsed, 'conf.client_id', None),
            'client_secret': getattr(parsed, 'conf.client_secret', None),
            'scope': getattr(parsed, 'conf.scope', None),
        }
        if getattr(parsed, 'description', None):
            kwargs['description'] = parsed.description
        try:
            token = api.Api().get_oauth2_token(**kwargs)
        except Exception as e:
            self.print_help(parser)
            cprint(
                'Error retrieving an OAuth2.0 token ({}).'.format(e.__class__),
                'red'
            )
        else:
            fmt = client.get_config('format')
            if fmt == 'human':
                print('export TOWER_OAUTH_TOKEN={}'.format(token))
            else:
                print(to_str(FORMATTERS[fmt]({'token': token}, '.')).strip())


class Config(CustomCommand):
    name = 'config'
    help_text = 'print current configuration values'

    def handle(self, client, parser):
        if client.help:
            parser.print_help()
            raise SystemExit()
        return {
            'base_url': config.base_url,
            'token': client.get_config('token'),
            'use_sessions': config.use_sessions,
            'credentials': config.credentials,
        }


class Import(CustomCommand):
    name = 'import'
    help_text = 'import resources into Tower'

    def handle(self, client, parser):
        if client.help:
            parser.print_help()
            raise SystemExit()

        fmt = client.get_config('format')
        if fmt == 'json':
            data = json.load(client.stdin)
        elif fmt == 'yaml':
            data = yaml.safe_load(client.stdin)
        else:
            raise ImportExportError("Unsupported format for Import: " + fmt)

        client.authenticate()
        client.v2.import_assets(data)

        return {}


class Export(CustomCommand):
    name = 'export'
    help_text = 'export resources from Tower'

    def extend_parser(self, parser):
        resources = parser.add_argument_group('resources')

        for resource in EXPORTABLE_RESOURCES:
            # This parsing pattern will result in 3 different possible outcomes:
            # 1) the resource flag is not used at all, which will result in the attr being None
            # 2) the resource flag is used with no argument, which will result in the attr being ''
            # 3) the resource flag is used with an argument, and the attr will be that argument's value
            resources.add_argument('--{}'.format(resource), nargs='?', const='')

    def handle(self, client, parser):
        self.extend_parser(parser)

        if client.help:
            parser.print_help()
            raise SystemExit()

        parsed = parser.parse_known_args()[0]
        kwargs = {resource: getattr(parsed, resource, None) for resource in EXPORTABLE_RESOURCES}

        client.authenticate()
        return client.v2.export_assets(**kwargs)


def parse_resource(client, skip_deprecated=False):
    subparsers = client.parser.add_subparsers(
        dest='resource',
        metavar='resource',
    )

    # check if the user is running a custom command
    for command in CustomCommand.__subclasses__():
        client.subparsers[command.name] = subparsers.add_parser(
            command.name, help=command.help_text
        )

    if hasattr(client, 'v2'):
        for k in client.v2.json.keys():
            if k in ('dashboard',):
                # the Dashboard API is deprecated and not supported
                continue

            # argparse aliases are *only* supported in Python3 (not 2.7)
            kwargs = {}
            if not skip_deprecated:
                if k in DEPRECATED_RESOURCES:
                    kwargs['aliases'] = [DEPRECATED_RESOURCES[k]]

            client.subparsers[k] = subparsers.add_parser(
                k, help='', **kwargs
            )

    resource = client.parser.parse_known_args()[0].resource
    if resource in DEPRECATED_RESOURCES.values():
        client.argv[
            client.argv.index(resource)
        ] = DEPRECATED_RESOURCES_REVERSE[resource]
        resource = DEPRECATED_RESOURCES_REVERSE[resource]

    if resource in CustomCommand.registry:
        parser = client.subparsers[resource]
        command = CustomCommand.registry[resource]()
        response = command.handle(client, parser)
        if response:
            _filter = client.get_config('filter')
            if (
                resource == 'config' and
                client.get_config('format') == 'human'
            ):
                response = {
                    'count': len(response),
                    'results': [
                        {'key': k, 'value': v}
                        for k, v in response.items()
                    ]
                }
                _filter = 'key, value'
            try:
                connection = client.root.connection
            except AttributeError:
                connection = None
            formatted = format_response(
                Page.from_json(response, connection=connection),
                fmt=client.get_config('format'),
                filter=_filter
            )
            print(formatted)
        raise SystemExit()
    else:
        return resource


def is_control_resource(resource):
    # special root level resources that don't don't represent database
    # entities that follow the list/detail semantic
    return resource in CONTROL_RESOURCES



# -*- coding: utf-8 -*-
from datetime import datetime
import sys

from unittest import mock
import pytest

from awxkit import utils
from awxkit import exceptions as exc


@pytest.mark.parametrize('inp, out',
                         [[True, True],
                          [False, False],
                          [1, True],
                          [0, False],
                          [1.0, True],
                          [0.0, False],
                          ['TrUe', True],
                          ['FalSe', False],
                          ['yEs', True],
                          ['No', False],
                          ['oN', True],
                          ['oFf', False],
                          ['asdf', True],
                          ['0', False],
                          ['', False],
                          [{1: 1}, True],
                          [{}, False],
                          [(0,), True],
                          [(), False],
                          [[1], True],
                          [[], False]])
def test_to_bool(inp, out):
    assert utils.to_bool(inp) == out


@pytest.mark.parametrize('inp, out',
                         [["{}", {}],
                          ["{'null': null}", {"null": None}],
                          ["{'bool': true}", {"bool": True}],
                          ["{'bool': false}", {"bool": False}],
                          ["{'int': 0}", {"int": 0}],
                          ["{'float': 1.0}", {"float": 1.0}],
                          ["{'str': 'abc'}", {"str": "abc"}],
                          ["{'obj': {}}", {"obj": {}}],
                          ["{'list': []}", {"list": []}],
                          ["---", None],
                          ["---\n'null': null", {'null': None}],
                          ["---\n'bool': true", {'bool': True}],
                          ["---\n'bool': false", {'bool': False}],
                          ["---\n'int': 0", {'int': 0}],
                          ["---\n'float': 1.0", {'float': 1.0}],
                          ["---\n'string': 'abc'", {'string': 'abc'}],
                          ["---\n'obj': {}", {'obj': {}}],
                          ["---\n'list': []", {'list': []}],
                          ["", None],
                          ["'null': null", {'null': None}],
                          ["'bool': true", {'bool': True}],
                          ["'bool': false", {'bool': False}],
                          ["'int': 0", {'int': 0}],
                          ["'float': 1.0", {'float': 1.0}],
                          ["'string': 'abc'", {'string': 'abc'}],
                          ["'obj': {}", {'obj': {}}],
                          ["'list': []", {'list': []}]])
def test_load_valid_json_or_yaml(inp, out):
    assert utils.load_json_or_yaml(inp) == out


@pytest.mark.parametrize('inp', [True, False, 0, 1.0, {}, [], None])
def test_load_invalid_json_or_yaml(inp):
    with pytest.raises(TypeError):
        utils.load_json_or_yaml(inp)


@pytest.mark.parametrize('non_ascii', [True, False])
@pytest.mark.skipif(
    sys.version_info < (3, 6),
    reason='this is only intended to be used in py3, not the CLI'
)
def test_random_titles_are_unicode(non_ascii):
    assert isinstance(utils.random_title(non_ascii=non_ascii), str)


@pytest.mark.parametrize('non_ascii', [True, False])
@pytest.mark.skipif(
    sys.version_info < (3, 6),
    reason='this is only intended to be used in py3, not the CLI'
)
def test_random_titles_generates_correct_characters(non_ascii):
    title = utils.random_title(non_ascii=non_ascii)
    if non_ascii:
        with pytest.raises(UnicodeEncodeError):
            title.encode('ascii')
        title.encode('utf-8')
    else:
        title.encode('ascii')
        title.encode('utf-8')


@pytest.mark.parametrize('inp, out',
                         [['ClassNameShouldChange', 'class_name_should_change'],
                          ['classnameshouldntchange', 'classnameshouldntchange'],
                          ['Classspacingshouldntchange', 'classspacingshouldntchange'],
                          ['Class1Name2Should3Change', 'class_1_name_2_should_3_change'],
                          ['Class123name234should345change456', 'class_123_name_234_should_345_change_456']])
def test_class_name_to_kw_arg(inp, out):
    assert utils.class_name_to_kw_arg(inp) == out


@pytest.mark.parametrize('first, second, expected',
                         [['/api/v2/resources/', '/api/v2/resources/', True],
                          ['/api/v2/resources/', '/api/v2/resources/?test=ignored', True],
                          ['/api/v2/resources/?one=ignored', '/api/v2/resources/?two=ignored', True],
                          ['http://one.com', 'http://one.com', True],
                          ['http://one.com', 'http://www.one.com', True],
                          ['http://one.com', 'http://one.com?test=ignored', True],
                          ['http://one.com', 'http://www.one.com?test=ignored', True],
                          ['http://one.com', 'https://one.com', False],
                          ['http://one.com', 'https://one.com?test=ignored', False]])
def test_are_same_endpoint(first, second, expected):
    assert utils.are_same_endpoint(first, second) == expected


@pytest.mark.parametrize('endpoint, expected',
                         [['/api/v2/resources/', 'v2'],
                          ['/api/v2000/resources/', 'v2000'],
                          ['/api/', 'common']])
def test_version_from_endpoint(endpoint, expected):
    assert utils.version_from_endpoint(endpoint) == expected


class OneClass:
    pass

class TwoClass:
    pass

class ThreeClass:
    pass

class FourClass(ThreeClass):
    pass

def test_filter_by_class_with_subclass_class():
    filtered = utils.filter_by_class((OneClass, OneClass), (FourClass, ThreeClass))
    assert filtered == [OneClass, FourClass]

def test_filter_by_class_with_subclass_instance():
    one = OneClass()
    four = FourClass()
    filtered = utils.filter_by_class((one, OneClass), (four, ThreeClass))
    assert filtered == [one, four]

def test_filter_by_class_no_arg_tuples():
    three = ThreeClass()
    filtered = utils.filter_by_class((True, OneClass), (False, TwoClass), (three, ThreeClass))
    assert filtered == [OneClass, None, three]

def test_filter_by_class_with_arg_tuples_containing_class():
    one = OneClass()
    three = (ThreeClass, dict(one=1, two=2))
    filtered = utils.filter_by_class((one, OneClass), (False, TwoClass), (three, ThreeClass))
    assert filtered == [one, None, three]

def test_filter_by_class_with_arg_tuples_containing_subclass():
    one = OneClass()
    three = (FourClass, dict(one=1, two=2))
    filtered = utils.filter_by_class((one, OneClass), (False, TwoClass), (three, ThreeClass))
    assert filtered == [one, None, three]

@pytest.mark.parametrize('truthy', (True, 123, 'yes'))
def test_filter_by_class_with_arg_tuples_containing_truthy(truthy):
    one = OneClass()
    three = (truthy, dict(one=1, two=2))
    filtered = utils.filter_by_class((one, OneClass), (False, TwoClass), (three, ThreeClass))
    assert filtered == [one, None, (ThreeClass, dict(one=1, two=2))]


@pytest.mark.parametrize('date_string,now,expected', [
    ('2017-12-20T00:00:01.5Z', datetime(2017, 12, 20, 0, 0, 2, 750000), 1.25),
    ('2017-12-20T00:00:01.5Z', datetime(2017, 12, 20, 0, 0, 1, 500000), 0.00),
    ('2017-12-20T00:00:01.5Z', datetime(2017, 12, 20, 0, 0, 0, 500000), -1.00),
])
def test_seconds_since_date_string(date_string, now, expected):
    with mock.patch('awxkit.utils.utcnow', return_value=now):
        assert utils.seconds_since_date_string(date_string) == expected


class RecordingCallback(object):

    def __init__(self, value=True):
        self.call_count = 0
        self.value = value

    def __call__(self):
        self.call_count += 1
        return self.value


def test_suppress():
    callback = RecordingCallback()

    with utils.suppress(ZeroDivisionError, IndexError):
        raise ZeroDivisionError
        callback()
        raise IndexError
        raise KeyError
    assert callback.call_count == 0

    with utils.suppress(ZeroDivisionError, IndexError):
        raise IndexError
        callback()
        raise ZeroDivisionError
        raise KeyError
    assert callback.call_count == 0

    with pytest.raises(KeyError):
        with utils.suppress(ZeroDivisionError, IndexError):
            raise KeyError
            callback()
            raise ZeroDivisionError
            raise IndexError
    assert callback.call_count == 0


class TestPollUntil(object):

    @pytest.mark.parametrize('timeout', [0, 0.0, -0.5, -1, -9999999])
    def test_callback_called_once_for_non_positive_timeout(self, timeout):
        with mock.patch('awxkit.utils.logged_sleep') as sleep:
            callback = RecordingCallback()
            utils.poll_until(callback, timeout=timeout)
            assert not sleep.called
            assert callback.call_count == 1

    def test_exc_raised_on_timeout(self):
        with mock.patch('awxkit.utils.logged_sleep'):
            with pytest.raises(exc.WaitUntilTimeout):
                utils.poll_until(lambda: False, timeout=0)

    @pytest.mark.parametrize('callback_value', [{'hello': 1}, 'foo', True])
    def test_non_falsey_callback_value_is_returned(self, callback_value):
        with mock.patch('awxkit.utils.logged_sleep'):
            assert utils.poll_until(lambda: callback_value) == callback_value


class TestPseudoNamespace(object):

    def test_set_item_check_item(self):
        pn = utils.PseudoNamespace()
        pn['key'] = 'value'
        assert pn['key'] == 'value'

    def test_set_item_check_attr(self):
        pn = utils.PseudoNamespace()
        pn['key'] = 'value'
        assert pn.key == 'value'

    def test_set_attr_check_item(self):
        pn = utils.PseudoNamespace()
        pn.key = 'value'
        assert pn['key'] == 'value'

    def test_set_attr_check_attr(self):
        pn = utils.PseudoNamespace()
        pn.key = 'value'
        assert pn.key == 'value'

    def test_auto_dicts_cast(self):
        pn = utils.PseudoNamespace()
        pn.one = dict()
        pn.one.two = dict(three=3)
        assert pn.one.two.three == 3
        assert pn == dict(one=dict(two=dict(three=3)))

    def test_auto_list_of_dicts_cast(self):
        pn = utils.PseudoNamespace()
        pn.one = [dict(two=2), dict(three=3)]
        assert pn.one[0].two == 2
        assert pn == dict(one=[dict(two=2), dict(three=3)])

    def test_auto_tuple_of_dicts_cast(self):
        pn = utils.PseudoNamespace()
        pn.one = (dict(two=2), dict(three=3))
        assert pn.one[0].two == 2
        assert pn == dict(one=(dict(two=2), dict(three=3)))

    def test_instantiation_via_dict(self):
        pn = utils.PseudoNamespace(dict(one=1, two=2, three=3))
        assert pn.one == 1
        assert pn == dict(one=1, two=2, three=3)
        assert len(pn.keys()) == 3

    def test_instantiation_via_kwargs(self):
        pn = utils.PseudoNamespace(one=1, two=2, three=3)
        assert pn.one == 1
        assert pn == dict(one=1, two=2, three=3)
        assert len(pn.keys()) == 3

    def test_instantiation_via_dict_and_kwargs(self):
        pn = utils.PseudoNamespace(dict(one=1, two=2, three=3), four=4, five=5)
        assert pn.one == 1
        assert pn.four == 4
        assert pn == dict(one=1, two=2, three=3, four=4, five=5)
        assert len(pn.keys()) == 5

    def test_instantiation_via_nested_dict(self):
        pn = utils.PseudoNamespace(dict(one=1, two=2), three=dict(four=4, five=dict(six=6)))
        assert pn.one == 1
        assert pn.three.four == 4
        assert pn.three.five.six == 6
        assert pn == dict(one=1, two=2, three=dict(four=4, five=dict(six=6)))

    def test_instantiation_via_nested_dict_with_list(self):
        pn = utils.PseudoNamespace(dict(one=[dict(two=2), dict(three=3)]))
        assert pn.one[0].two == 2
        assert pn.one[1].three == 3
        assert pn == dict(one=[dict(two=2), dict(three=3)])

    def test_instantiation_via_nested_dict_with_lists(self):
        pn = utils.PseudoNamespace(dict(one=[dict(two=2),
                                             dict(three=dict(four=4,
                                                             five=[dict(six=6),
                                                                   dict(seven=7)]))]))
        assert pn.one[1].three.five[1].seven == 7

    def test_instantiation_via_nested_dict_with_tuple(self):
        pn = utils.PseudoNamespace(dict(one=(dict(two=2), dict(three=3))))
        assert pn.one[0].two == 2
        assert pn.one[1].three == 3
        assert pn == dict(one=(dict(two=2), dict(three=3)))

    def test_instantiation_via_nested_dict_with_tuples(self):
        pn = utils.PseudoNamespace(dict(one=(dict(two=2),
                                             dict(three=dict(four=4,
                                                             five=(dict(six=6),
                                                                   dict(seven=7)))))))
        assert pn.one[1].three.five[1].seven == 7

    def test_update_with_nested_dict(self):
        pn = utils.PseudoNamespace()
        pn.update(dict(one=1, two=2, three=3), four=4, five=5)
        assert pn.one == 1
        assert pn.four == 4
        assert pn == dict(one=1, two=2, three=3, four=4, five=5)
        assert len(pn.keys()) == 5

    def test_update_with_nested_dict_with_lists(self):
        pn = utils.PseudoNamespace()
        pn.update(dict(one=[dict(two=2),
                            dict(three=dict(four=4,
                                            five=[dict(six=6),
                                                  dict(seven=7)]))]))
        assert pn.one[1].three.five[1].seven == 7

    def test_update_with_nested_dict_with_tuples(self):
        pn = utils.PseudoNamespace()
        pn.update(dict(one=(dict(two=2),
                            dict(three=dict(four=4,
                                            five=(dict(six=6),
                                                  dict(seven=7)))))))
        assert pn.one[1].three.five[1].seven == 7


class TestUpdatePayload(object):

    def test_empty_payload(self):
        fields = ('one', 'two', 'three', 'four')
        kwargs = dict(two=2, four=4)
        payload = {}
        utils.update_payload(payload, fields, kwargs)
        assert payload == kwargs

    def test_untouched_payload(self):
        fields = ('not', 'in', 'kwargs')
        kwargs = dict(one=1, two=2)
        payload = dict(three=3, four=4)
        utils.update_payload(payload, fields, kwargs)
        assert payload == dict(three=3, four=4)

    def test_overwritten_payload(self):
        fields = ('one', 'two')
        kwargs = dict(one=1, two=2)
        payload = dict(one='one', two='two')
        utils.update_payload(payload, fields, kwargs)
        assert payload == kwargs

    def test_falsy_kwargs(self):
        fields = ('one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight')
        kwargs = dict(one=False, two=(), three='', four=None, five=0, six={}, seven=set(), eight=[])
        payload = {}
        utils.update_payload(payload, fields, kwargs)
        assert payload == kwargs

    def test_not_provided_strips_payload(self):
        fields = ('one', 'two')
        kwargs = dict(one=utils.not_provided)
        payload = dict(one=1, two=2)
        utils.update_payload(payload, fields, kwargs)
        assert payload == dict(two=2)


def test_to_ical():
    now = datetime.utcnow()
    ical_datetime = utils.to_ical(now)
    date = str(now.date()).replace('-', '')
    time = str(now.time()).split('.')[0].replace(':', '')
    assert ical_datetime == '{}T{}Z'.format(date, time)



from awxkit.api.mixins import HasCreate, DSAdapter
from awxkit.utils import random_title, PseudoNamespace
from awxkit.api.resources import resources
from awxkit.config import config

from . import base
from . import page


class User(HasCreate, base.Base):

    NATURAL_KEY = ('username',)

    def payload(self, **kwargs):
        payload = PseudoNamespace(
            username=kwargs.get('username') or 'User-{}'.format(
                random_title(
                    non_ascii=False)),
            password=kwargs.get('password') or config.credentials.default.password,
            is_superuser=kwargs.get(
                'is_superuser',
                False),
            is_system_auditor=kwargs.get(
                'is_system_auditor',
                False),
            first_name=kwargs.get(
                'first_name',
                random_title()),
            last_name=kwargs.get(
                'last_name',
                random_title()),
            email=kwargs.get(
                'email',
                '{}@example.com'.format(random_title(5, non_ascii=False))
            )
        )
        return payload

    def create_payload(self, username='', password='', **kwargs):
        payload = self.payload(username=username, password=password, **kwargs)
        payload.ds = DSAdapter(self.__class__.__name__, self._dependency_store)
        return payload

    def create(self, username='', password='', organization=None, **kwargs):
        payload = self.create_payload(
            username=username, password=password, **kwargs)
        self.password = payload.password

        self.update_identity(Users(self.connection).post(payload))

        if organization:
            organization.add_user(self)

        return self


page.register_page([resources.user,
                    (resources.users, 'post')], User)


class Users(page.PageList, User):

    pass


page.register_page([resources.users,
                    resources.organization_admins,
                    resources.related_users,
                    resources.credential_owner_users,
                    resources.user_admin_organizations], Users)


class Me(Users):

    pass


page.register_page(resources.me, Me)



from . import base
from . import page

from awxkit.api.resources import resources


class SurveySpec(base.Base):

    def get_variable_default(self, var):
        for item in self.spec:
            if item.get('variable') == var:
                return item.get('default')

    def get_default_vars(self):
        default_vars = dict()
        for item in self.spec:
            if item.get("default", None):
                default_vars[item.variable] = item.default
        return default_vars

    def get_required_vars(self):
        required_vars = []
        for item in self.spec:
            if item.get("required", None):
                required_vars.append(item.variable)
        return required_vars


page.register_page([resources.job_template_survey_spec,
                    resources.workflow_job_template_survey_spec], SurveySpec)



import pytest

from awxkit.api.registry import URLRegistry


class One(object):

    pass


class Two(object):

    pass


@pytest.fixture
def reg():
    return URLRegistry()


def test_url_pattern(reg):
    desired = r'^/some/resources/\d+/(\?.*)*$'
    assert reg.url_pattern(r'/some/resources/\d+/').pattern == desired


def test_methodless_get_from_empty_registry(reg):
    assert reg.get('nonexistent') is None


def test_method_get_from_empty_registry(reg):
    assert reg.get('nonexistent', 'method') is None


def test_methodless_setdefault_methodless_get(reg):
    reg.setdefault(One)
    assert reg.get('some_path') is One


def test_methodless_setdefault_method_get(reg):
    reg.setdefault(One)
    assert reg.get('some_path', 'method') is One


def test_method_setdefault_methodless_get(reg):
    reg.setdefault('method', One)
    assert reg.get('some_path') is None


def test_method_setdefault_matching_method_get(reg):
    reg.setdefault('method', One)
    assert reg.get('some_path', 'method') is One


def test_method_setdefault_nonmatching_method_get(reg):
    reg.setdefault('method', One)
    assert reg.get('some_path', 'nonexistent') is None


def test_multimethod_setdefault_matching_method_get(reg):
    reg.setdefault(('method_one', 'method_two'), One)
    assert reg.get('some_path', 'method_one') is One
    assert reg.get('some_path', 'method_two') is One


def test_multimethod_setdefault_nonmatching_method_get(reg):
    reg.setdefault(('method_one', 'method_two'), One)
    assert reg.get('some_path') is None
    assert reg.get('some_path', 'nonexistent') is None


def test_wildcard_setdefault_methodless_get(reg):
    reg.setdefault('.*', One)
    assert reg.get('some_path') is One


def test_wildcard_setdefault_method_get(reg):
    reg.setdefault('.*', One)
    assert reg.get('some_path', 'method') is One


def test_regex_method_setdefaults_over_wildcard_method_get(reg):
    reg.setdefault('.*', One)
    reg.setdefault('reg.*ex', Two)
    for _ in range(1000):
        assert reg.get('some_path', 'regex') is Two


def test_methodless_registration_with_matching_path_methodless_get(reg):
    reg.register('some_path', One)
    assert reg.get('some_path') is One


def test_methodless_registraion_with_nonmatching_path_methodless_get(reg):
    reg.register('some_path', One)
    assert reg.get('nonexistent') is None


def test_methodless_registration_with_matching_path_nonmatching_method_get(reg):
    reg.register('some_path', One)
    assert reg.get('some_path', 'method') is None


def test_method_registration_with_matching_path_matching_method_get(reg):
    reg.register('some_path', 'method', One)
    assert reg.get('some_path', 'method') is One


def test_method_registration_with_matching_path_nonmatching_method_get(reg):
    reg.register('some_path', 'method_one', One)
    assert reg.get('some_path', 'method_two') is None


def test_multimethod_registration_with_matching_path_matching_method_get(reg):
    reg.register('some_path', ('method_one', 'method_two'), One)
    assert reg.get('some_path', 'method_one') is One
    assert reg.get('some_path', 'method_two') is One


def test_multimethod_registration_with_path_matching_method_get(reg):
    reg.register('some_path', ('method_one', 'method_two'), One)
    assert reg.get('some_path', 'method_three') is None


def test_multipath_methodless_registration_with_matching_path_methodless_get(reg):
    reg.register(('some_path_one', 'some_path_two'), One)
    assert reg.get('some_path_one') is One
    assert reg.get('some_path_two') is One


def test_multipath_methodless_registration_with_matching_path_nonmatching_method_get(reg):
    reg.register(('some_path_one', 'some_path_two'), One)
    assert reg.get('some_path_one', 'method') is None
    assert reg.get('some_path_two', 'method') is None


def test_multipath_method_registration_with_matching_path_matching_method_get(reg):
    reg.register((('some_path_one', 'method_one'), ('some_path_two', 'method_two')), One)
    assert reg.get('some_path_one', 'method_one') is One
    assert reg.get('some_path_two', 'method_two') is One


def test_multipath_partial_method_registration_with_matching_path_matching_method_get(reg):
    reg.register(('some_path_one', ('some_path_two', 'method')), One)
    assert reg.get('some_path_one') is One
    assert reg.get('some_path_two', 'method') is One


def test_wildcard_method_registration_with_methodless_get(reg):
    reg.register('some_path', '.*', One)
    assert reg.get('some_path') is One


def test_wildcard_method_registration_with_method_get(reg):
    reg.register('some_path', '.*', One)
    assert reg.get('some_path', 'method') is One


def test_wildcard_and_specific_method_registration_acts_as_default(reg):
    reg.register('some_path', 'method_one', Two)
    reg.register('some_path', '.*', One)
    reg.register('some_path', 'method_two', Two)
    for _ in range(1000):  # eliminate overt randomness
        assert reg.get('some_path', 'nonexistent') is One
        assert reg.get('some_path', 'method_one') is Two
        assert reg.get('some_path', 'method_two') is Two


@pytest.mark.parametrize('method', ('method', '.*'))
def test_multiple_method_registrations_disallowed_for_single_path_single_registration(reg, method):
    with pytest.raises(TypeError) as e:
        reg.register((('some_path', method), ('some_path', method)), One)
    assert str(e.value) == ('"{0.pattern}" already has registered method "{1}"'
                          .format(reg.url_pattern('some_path'), method))


@pytest.mark.parametrize('method', ('method', '.*'))
def test_multiple_method_registrations_disallowed_for_single_path_multiple_registrations(reg, method):
    reg.register('some_path', method, One)
    with pytest.raises(TypeError) as e:
        reg.register('some_path', method, One)
    assert str(e.value) == ('"{0.pattern}" already has registered method "{1}"'
                          .format(reg.url_pattern('some_path'), method))


def test_paths_can_be_patterns(reg):
    reg.register('.*pattern.*', One)
    assert reg.get('XYZpattern123') is One


def test_mixed_form_single_registration(reg):
    reg.register([('some_path_one', 'method_one'),
                  'some_path_two',
                  ('some_path_three', ('method_two', 'method_three')),
                  'some_path_four', 'some_path_five'], One)
    assert reg.get('some_path_one', 'method_one') is One
    assert reg.get('some_path_one') is None
    assert reg.get('some_path_one', 'nonexistent') is None
    assert reg.get('some_path_two') is One
    assert reg.get('some_path_two', 'nonexistent') is None
    assert reg.get('some_path_three', 'method_two') is One
    assert reg.get('some_path_three', 'method_three') is One
    assert reg.get('some_path_three') is None
    assert reg.get('some_path_three', 'nonexistent') is None
    assert reg.get('some_path_four') is One
    assert reg.get('some_path_four', 'nonexistent') is None
    assert reg.get('some_path_five') is One
    assert reg.get('some_path_five', 'nonexistent') is None


def test_mixed_form_single_registration_with_methodless_default(reg):
    reg.setdefault(One)
    reg.register([('some_path_one', 'method_one'),
                  'some_path_two',
                  ('some_path_three', ('method_two', 'method_three')),
                  'some_path_four', 'some_path_five'], Two)
    assert reg.get('some_path_one', 'method_one') is Two
    assert reg.get('some_path_one') is One
    assert reg.get('some_path_one', 'nonexistent') is One
    assert reg.get('some_path_two') is Two
    assert reg.get('some_path_two', 'nonexistent') is One
    assert reg.get('some_path_three', 'method_two') is Two
    assert reg.get('some_path_three', 'method_three') is Two
    assert reg.get('some_path_three') is One
    assert reg.get('some_path_three', 'nonexistent') is One
    assert reg.get('some_path_four') is Two
    assert reg.get('some_path_four', 'nonexistent') is One
    assert reg.get('some_path_five') is Two
    assert reg.get('some_path_five', 'nonexistent') is One


def test_mixed_form_single_registration_with_method_default(reg):
    reg.setdefault('existent', One)
    reg.register([('some_path_one', 'method_one'),
                  'some_path_two',
                  ('some_path_three', ('method_two', 'method_three')),
                  'some_path_four', 'some_path_five'], Two)
    assert reg.get('some_path_one', 'method_one') is Two
    assert reg.get('some_path_one') is None
    assert reg.get('some_path_one', 'existent') is One
    assert reg.get('some_path_one', 'nonexistent') is None
    assert reg.get('some_path_two') is Two
    assert reg.get('some_path_two', 'existent') is One
    assert reg.get('some_path_two', 'nonexistent') is None
    assert reg.get('some_path_three', 'method_two') is Two
    assert reg.get('some_path_three', 'method_three') is Two
    assert reg.get('some_path_three') is None
    assert reg.get('some_path_three', 'existent') is One
    assert reg.get('some_path_three', 'nonexistent') is None
    assert reg.get('some_path_four') is Two
    assert reg.get('some_path_four', 'existent') is One
    assert reg.get('some_path_four', 'nonexistent') is None
    assert reg.get('some_path_five') is Two
    assert reg.get('some_path_five', 'existent') is One
    assert reg.get('some_path_five', 'nonexistent') is None



from awxkit.utils import suppress
import awxkit.exceptions as exc


notification_endpoints = ("notification_templates", "notification_templates_started", "notification_templates_error",
                          "notification_templates_success")
wfjt_notification_endpoints = notification_endpoints + ('notification_templates_approvals',)


class HasNotifications(object):

    def add_notification_template(self, notification_template, endpoint="notification_templates_success"):
        from awxkit.api.pages.workflow_job_templates import WorkflowJobTemplate
        supported_endpoints = wfjt_notification_endpoints if isinstance(self, WorkflowJobTemplate) \
            else notification_endpoints
        if endpoint not in supported_endpoints:
            raise ValueError('Unsupported notification endpoint "{0}". Please use one of {1}.'
                             .format(endpoint, notification_endpoints))
        with suppress(exc.NoContent):
            self.related[endpoint].post(dict(id=notification_template.id))

    def remove_notification_template(self, notification_template, endpoint="notification_templates_success"):
        from awxkit.api.pages.workflow_job_templates import WorkflowJobTemplate
        supported_endpoints = wfjt_notification_endpoints if isinstance(self, WorkflowJobTemplate) \
            else notification_endpoints
        if endpoint not in supported_endpoints:
            raise ValueError('Unsupported notification endpoint "{0}". Please use one of {1}.'
                             .format(endpoint, notification_endpoints))
        with suppress(exc.NoContent):
            self.related[endpoint].post(dict(id=notification_template.id, disassociate=notification_template.id))



from __future__ import print_function

import logging
import os
import pkg_resources
import sys

from requests.exceptions import RequestException

from .custom import handle_custom_actions
from .format import (add_authentication_arguments,
                     add_output_formatting_arguments,
                     FORMATTERS, format_response)
from .options import ResourceOptionsParser, UNIQUENESS_RULES
from .resource import parse_resource, is_control_resource
from awxkit import api, config, utils, exceptions, WSClient  # noqa
from awxkit.cli.utils import HelpfulArgumentParser, cprint, disable_color, colored
from awxkit.awx.utils import uses_sessions  # noqa


__version__ = pkg_resources.get_distribution('awxkit').version


class CLI(object):
    """A programmatic HTTP OPTIONS-based CLI for AWX/Ansible Tower.

    This CLI works by:

    - Configuring CLI options via Python's argparse (authentication, formatting
      options, etc...)
    - Discovering AWX API endpoints at /api/v2/ and mapping them to _resources_
    - Discovering HTTP OPTIONS _actions_ on resources to determine how
      resources can be interacted with (e.g., list, modify, delete, etc...)
    - Parsing sys.argv to map CLI arguments and flags to
      awxkit SDK calls

    ~ awx <resource> <action> --parameters

    e.g.,

    ~ awx users list -v
    GET /api/ HTTP/1.1" 200
    GET /api/v2/ HTTP/1.1" 200
    POST /api/login/ HTTP/1.1" 302
    OPTIONS /api/v2/users/ HTTP/1.1" 200
    GET /api/v2/users/
    {
     "count": 2,
     "results": [
     ...

    Interacting with this class generally involves a few critical methods:

    1.  parse_args() - this method is used to configure and parse global CLI
        flags, such as formatting flags, and arguments which represent client
        configuration (including authentication details)
    2.  connect() - once configuration is parsed, this method fetches /api/v2/
        and itemizes the list of supported resources
    3.  parse_resource() - attempts to parse the <resource> specified on the
        command line (e.g., users, organizations), including logic
        for discovering available actions for endpoints using HTTP OPTIONS
        requests

    At multiple stages of this process, an internal argparse.ArgumentParser()
    is progressively built and parsed based on sys.argv, (meaning, that if you
    supply invalid or incomplete arguments, argparse will print the usage
    message and an explanation of what you got wrong).
    """

    subparsers = {}
    original_action = None

    def __init__(self, stdout=sys.stdout, stderr=sys.stderr, stdin=sys.stdin):
        self.stdout = stdout
        self.stderr = stderr
        self.stdin = stdin

    def get_config(self, key):
        """Helper method for looking up the value of a --conf.xyz flag"""
        return getattr(self.args, 'conf.{}'.format(key))

    @property
    def help(self):
        return '--help' in self.argv or '-h' in self.argv

    def authenticate(self):
        """Configure the current session (or OAuth2.0 token)"""
        token = self.get_config('token')
        if token:
            self.root.connection.login(
                None, None, token=token,
            )
        else:
            config.use_sessions = True
            self.root.load_session().get()

    def connect(self):
        """Fetch top-level resources from /api/v2"""
        config.base_url = self.get_config('host')
        config.client_connection_attempts = 1
        config.assume_untrusted = False
        if self.get_config('insecure'):
            config.assume_untrusted = True

        config.credentials = utils.PseudoNamespace({
            'default': {
                'username': self.get_config('username'),
                'password': self.get_config('password'),
            }
        })

        _, remainder = self.parser.parse_known_args()
        if remainder and remainder[0] == 'config':
            # the config command is special; it doesn't require
            # API connectivity
            return
        # ...otherwise, set up a awxkit connection because we're
        # likely about to do some requests to /api/v2/
        self.root = api.Api()
        try:
            self.fetch_version_root()
        except RequestException:
            # If we can't reach the API root (this usually means that the
            # hostname is wrong, or the credentials are wrong)
            if self.help:
                # ...but the user specified -h...
                known, unknown = self.parser.parse_known_args(self.argv)
                if len(unknown) == 1 and os.path.basename(unknown[0]) == 'awx':
                    return
            raise

    def fetch_version_root(self):
        try:
            self.v2 = self.root.get().available_versions.v2.get()
        except AttributeError:
            raise RuntimeError(
                'An error occurred while fetching {}/api/'.format(
                    self.get_config('host')
                )
            )

    def parse_resource(self, skip_deprecated=False):
        """Attempt to parse the <resource> (e.g., jobs) specified on the CLI

        If a valid resource is discovered, the user will be authenticated
        (either via an OAuth2.0 token or session-based auth) and the remaining
        CLI arguments will be processed (to determine the requested action
        e.g., list, create, delete)

        :param skip_deprecated: when False (the default), deprecated resource
                                names from the open source tower-cli project
                                will be allowed
        """
        self.resource = parse_resource(self, skip_deprecated=skip_deprecated)
        if self.resource:
            self.authenticate()
            resource = getattr(self.v2, self.resource)
            if is_control_resource(self.resource):
                # control resources are special endpoints that you can only
                # do an HTTP GET to, and which return plain JSON metadata
                # examples are `/api/v2/ping/`, `/api/v2/config/`, etc...
                if self.help:
                    self.subparsers[self.resource].print_help()
                    raise SystemExit()
                self.method = 'get'
                response = getattr(resource, self.method)()
            else:
                response = self.parse_action(resource)

            _filter = self.get_config('filter')

            # human format for metrics, settings is special
            if (
                self.resource in ('metrics', 'settings') and
                self.get_config('format') == 'human'
            ):
                response.json = {
                    'count': len(response.json),
                    'results': [
                        {'key': k, 'value': v}
                        for k, v in response.json.items()
                    ]
                }
                _filter = 'key, value'

            if (
                self.get_config('format') == 'human' and
                _filter == '.' and
                self.resource in UNIQUENESS_RULES
            ):
                _filter = ', '.join(UNIQUENESS_RULES[self.resource])

            formatted = format_response(
                response,
                fmt=self.get_config('format'),
                filter=_filter,
                changed=self.original_action in (
                    'modify', 'create', 'associate', 'disassociate'
                )
            )
            if formatted:
                print(utils.to_str(formatted), file=self.stdout)
            if hasattr(response, 'rc'):
                raise SystemExit(response.rc)
        else:
            self.parser.print_help()

    def parse_action(self, page, from_sphinx=False):
        """Perform an HTTP OPTIONS request

        This method performs an HTTP OPTIONS request to build a list of valid
        actions, and (if provided) runs the code for the action specified on
        the CLI

        :param page: a awxkit.api.pages.TentativePage object representing the
                     top-level resource in question (e.g., /api/v2/jobs)
        :param from_sphinx: a flag specified by our sphinx plugin, which allows
                            us to walk API OPTIONS using this function
                            _without_ triggering a SystemExit (argparse's
                            behavior if required arguments are missing)
        """
        subparsers = self.subparsers[self.resource].add_subparsers(
            dest='action',
            metavar='action'
        )
        subparsers.required = True

        # parse the action from OPTIONS
        parser = ResourceOptionsParser(self.v2, page, self.resource, subparsers)
        if parser.deprecated:
            description = 'This resource has been deprecated and will be removed in a future release.'
            if not from_sphinx:
                description = colored(description, 'yellow')
            self.subparsers[self.resource].description = description

        if from_sphinx:
            # Our Sphinx plugin runs `parse_action` for *every* available
            # resource + action in the API so that it can generate usage
            # strings for automatic doc generation.
            #
            # Because of this behavior, we want to silently ignore the
            # `SystemExit` argparse will raise when you're missing required
            # positional arguments (which some actions have).
            try:
                self.parser.parse_known_args(self.argv)[0]
            except SystemExit:
                pass
        else:
            self.parser.parse_known_args()[0]

        # parse any action arguments
        if self.resource != 'settings':
            for method in ('list', 'modify', 'create'):
                if method in parser.parser.choices:
                    parser.build_query_arguments(
                        method,
                        'GET' if method == 'list' else 'POST'
                    )
        if from_sphinx:
            parsed, extra = self.parser.parse_known_args(self.argv)
        else:
            parsed, extra = self.parser.parse_known_args()

        if extra and self.verbose:
            # If extraneous arguments were provided, warn the user
            cprint('{}: unrecognized arguments: {}'.format(
                self.parser.prog,
                ' '.join(extra)
            ), 'yellow', file=self.stdout)

        # build a dictionary of all of the _valid_ flags specified on the
        # command line so we can pass them on to the underlying awxkit call
        # we ignore special global flags like `--help` and `--conf.xyz`, and
        # the positional resource argument (i.e., "jobs")
        # everything else is a flag used as a query argument for the HTTP
        # request we'll make (e.g., --username="Joe", --verbosity=3)
        parsed = parsed.__dict__
        parsed = dict(
            (k, v) for k, v in parsed.items()
            if (
                v is not None and
                k not in ('help', 'resource') and
                not k.startswith('conf.')
            )
        )

        # if `id` is one of the arguments, it's a detail view
        if 'id' in parsed:
            page.endpoint += '{}/'.format(str(parsed.pop('id')))

        # determine the awxkit method to call
        action = self.original_action = parsed.pop('action')
        page, action = handle_custom_actions(
            self.resource, action, page
        )
        self.method = {
            'list': 'get',
            'modify': 'patch',
        }.get(action, action)

        if self.method == 'patch' and not parsed:
            # If we're doing an HTTP PATCH with an empty payload,
            # just print the help message (it's a no-op anyways)
            parser.parser.choices['modify'].print_help()
            return

        if self.help:
            # If --help is specified on a subarg parser, bail out
            # and print its help text
            parser.parser.choices[self.original_action].print_help()
            return

        if self.original_action == 'create':
            return page.post(parsed)

        return getattr(page, self.method)(**parsed)

    def parse_args(self, argv, env=None):
        """Configure the global parser.ArgumentParser object and apply
        global flags (such as --help, authentication, and formatting arguments)
        """
        env = env or os.environ
        self.argv = argv
        self.parser = HelpfulArgumentParser(add_help=False)
        self.parser.add_argument(
            '--help',
            action='store_true',
            help='prints usage information for the awx tool',
        )
        self.parser.add_argument(
            '--version',
            dest='conf.version',
            action='version',
            help='display awx CLI version',
            version=__version__
        )
        add_authentication_arguments(self.parser, env)
        add_output_formatting_arguments(self.parser, env)

        self.args = self.parser.parse_known_args(self.argv)[0]
        self.verbose = self.get_config('verbose')
        if self.verbose:
            logging.basicConfig(level='DEBUG')
        self.color = self.get_config('color')
        if not self.color:
            disable_color()
        fmt = self.get_config('format')
        if fmt not in FORMATTERS.keys():
            self.parser.error('No formatter %s available.' % (fmt))



import json
import sys
import traceback
import yaml
import urllib3

from requests.exceptions import ConnectionError, SSLError

from .client import CLI
from awxkit.utils import to_str
from awxkit.exceptions import Unauthorized, Common
from awxkit.cli.utils import cprint


# you'll only see these warnings if you've explicitly *disabled* SSL
# verification, so they're a little annoying, redundant
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def run(stdout=sys.stdout, stderr=sys.stderr, argv=[]):
    cli = CLI(stdout=stdout, stderr=stderr)
    try:
        cli.parse_args(argv or sys.argv)
        cli.connect()
        cli.parse_resource()
    except KeyboardInterrupt:
        sys.exit(1)
    except ConnectionError as e:
        cli.parser.print_help()
        msg = (
            '\nThere was a network error of some kind trying to reach '
            '{}.\nYou might need to specify (or double-check) '
            '--conf.host'.format(cli.get_config('host'))
        )
        if isinstance(e, SSLError):
            msg = (
                '\nCould not establish a secure connection.  '
                '\nPlease add your server to your certificate authority.'
                '\nYou can also run this command by specifying '
                '-k or --conf.insecure'
            )
        cprint(msg + '\n', 'red', file=stderr)
        cprint(e, 'red', file=stderr)
        sys.exit(1)
    except Unauthorized as e:
        cli.parser.print_help()
        msg = '\nValid credentials were not provided.\n$ awx login --help'
        cprint(msg + '\n', 'red', file=stderr)
        if cli.verbose:
            cprint(e.__class__, 'red', file=stderr)
        sys.exit(1)
    except Common as e:
        if cli.verbose:
            print(traceback.format_exc(), sys.stderr)
        if cli.get_config('format') == 'json':
            json.dump(e.msg, sys.stdout)
            print('')
        elif cli.get_config('format') == 'yaml':
            sys.stdout.write(to_str(
                yaml.safe_dump(
                    e.msg,
                    default_flow_style=False,
                    encoding='utf-8',
                    allow_unicode=True
                )
            ))
        elif cli.get_config('format') == 'human':
            sys.stdout.write(e.__class__.__name__)
            print('')
        sys.exit(1)
    except Exception as e:
        if cli.verbose:
            e = traceback.format_exc()
        cprint(e, 'red', file=stderr)
        sys.exit(1)



from awxkit.api.resources import resources
from awxkit.utils import random_title, update_payload
from awxkit.api.mixins import HasStatus
from . import base
from . import page


class UnifiedJobTemplate(HasStatus, base.Base):
    """Base class for unified job template pages (e.g. project, inventory_source,
    and job_template).
    """

    optional_schedule_fields = (
        'extra_data',
        'diff_mode',
        'limit',
        'job_tags',
        'skip_tags',
        'job_type',
        'verbosity',
        'inventory',
    )

    def __str__(self):
        # NOTE: I use .replace('%', '%%') to workaround an odd string
        # formatting issue where result_stdout contained '%s'.  This later caused
        # a python traceback when attempting to display output from this
        # method.
        items = [
            'id',
            'name',
            'status',
            'source',
            'last_update_failed',
            'last_updated',
            'result_traceback',
            'job_explanation',
            'job_args']
        info = []
        for item in [x for x in items if hasattr(self, x)]:
            info.append('{0}:{1}'.format(item, getattr(self, item)))
        output = '<{0.__class__.__name__} {1}>'.format(self, ', '.join(info))
        return output.replace('%', '%%')

    def add_schedule(
            self,
            name='',
            description='',
            enabled=True,
            rrule=None,
            **kwargs):
        if rrule is None:
            rrule = "DTSTART:30180101T000000Z RRULE:FREQ=YEARLY;INTERVAL=1"
        payload = dict(
            name=name or "{0} Schedule {1}".format(
                self.name,
                random_title()),
            description=description or random_title(10),
            enabled=enabled,
            rrule=str(rrule))

        update_payload(payload, self.optional_schedule_fields, kwargs)

        return self.related.schedules.post(payload)

    @property
    def is_successful(self):
        """An unified_job_template is considered successful when:
        1) status == 'successful'
        2) not last_update_failed
        3) last_updated
        """
        return super(
            UnifiedJobTemplate,
            self).is_successful and not self.last_update_failed and self.last_updated is not None


page.register_page(resources.unified_job_template, UnifiedJobTemplate)


class UnifiedJobTemplates(page.PageList, UnifiedJobTemplate):

    pass


page.register_page(resources.unified_job_templates, UnifiedJobTemplates)



from awxkit.api.resources import resources
from . import base
from . import page


class Setting(base.Base):

    pass


page.register_page([resources.setting,
                    resources.settings_all,
                    resources.settings_authentication,
                    resources.settings_changed,
                    resources.settings_github,
                    resources.settings_github_org,
                    resources.settings_github_team,
                    resources.settings_google_oauth2,
                    resources.settings_jobs,
                    resources.settings_ldap,
                    resources.settings_radius,
                    resources.settings_saml,
                    resources.settings_system,
                    resources.settings_tacacsplus,
                    resources.settings_ui,
                    resources.settings_user,
                    resources.settings_user_defaults], Setting)


class Settings(page.PageList, Setting):

    def get_endpoint(self, endpoint):
        """Helper method used to navigate to a specific settings endpoint.
        (Pdb) settings_pg.get_endpoint('all')
        """
        base_url = '{0}{1}/'.format(self.endpoint, endpoint)
        return self.walk(base_url)

    get_setting = get_endpoint


page.register_page(resources.settings, Settings)



from awxkit.utils import suppress
import awxkit.exceptions as exc


class HasInstanceGroups(object):

    def add_instance_group(self, instance_group):
        with suppress(exc.NoContent):
            self.related['instance_groups'].post(dict(id=instance_group.id))

    def remove_instance_group(self, instance_group):
        with suppress(exc.NoContent):
            self.related['instance_groups'].post(dict(id=instance_group.id, disassociate=instance_group.id))

    def remove_all_instance_groups(self):
        for ig in self.related.instance_groups.get().results:
            self.remove_instance_group(ig)



from awxkit.api.pages import UnifiedJob
from awxkit.api.resources import resources
from . import page


class WorkflowJob(UnifiedJob):

    def __str__(self):
        # TODO: Update after endpoint's fields are finished filling out
        return super(UnifiedJob, self).__str__()

    def relaunch(self, payload={}):
        result = self.related.relaunch.post(payload)
        return self.walk(result.url)

    @property
    def result_stdout(self):
        # workflow jobs do not have result_stdout
        # which is problematic for the UnifiedJob.is_successful reliance on
        # related stdout endpoint.
        if 'result_stdout' not in self.json:
            return 'Unprovided AWX field.'
        else:
            return super(WorkflowJob, self).result_stdout


page.register_page(resources.workflow_job, WorkflowJob)


class WorkflowJobs(page.PageList, WorkflowJob):

    pass


page.register_page([resources.workflow_jobs,
                    resources.workflow_job_template_jobs,
                    resources.job_template_slice_workflow_jobs],
                   WorkflowJobs)



from awxkit.utils import random_title, PseudoNamespace
from awxkit.api.mixins import HasCreate, DSAdapter
from awxkit.api.resources import resources
from awxkit.api.pages import Organization
from . import base
from . import page


class Label(HasCreate, base.Base):

    dependencies = [Organization]
    NATURAL_KEY = ('organization', 'name')

    def silent_delete(self):
        """Label pages do not support DELETE requests. Here, we override the base page object
        silent_delete method to account for this.
        """
        pass

    def payload(self, organization, **kwargs):
        payload = PseudoNamespace(
            name=kwargs.get('name') or 'Label - {}'.format(
                random_title()),
            description=kwargs.get('description') or random_title(10),
            organization=organization.id)
        return payload

    def create_payload(
            self,
            name='',
            description='',
            organization=Organization,
            **kwargs):
        self.create_and_update_dependencies(organization)
        payload = self.payload(
            organization=self.ds.organization,
            name=name,
            description=description,
            **kwargs)
        payload.ds = DSAdapter(self.__class__.__name__, self._dependency_store)
        return payload

    def create(
            self,
            name='',
            description='',
            organization=Organization,
            **kwargs):
        payload = self.create_payload(
            name=name,
            description=description,
            organization=organization,
            **kwargs)
        return self.update_identity(Labels(self.connection).post(payload))


page.register_page([resources.label,
                    (resources.labels, 'post')], Label)


class Labels(page.PageList, Label):

    pass


page.register_page([resources.labels,
                    resources.job_labels,
                    resources.job_template_labels,
                    resources.workflow_job_template_labels], Labels)



import pytest
from requests.exceptions import ConnectionError

from awxkit.cli import CLI
from awxkit import config

def test_host_from_environment():
    cli = CLI()
    cli.parse_args(
        ['awx'],
        env={'TOWER_HOST': 'https://xyz.local'}
    )
    with pytest.raises(ConnectionError):
        cli.connect()
    assert config.base_url == 'https://xyz.local'

def test_host_from_argv():
    cli = CLI()
    cli.parse_args(['awx', '--conf.host', 'https://xyz.local'])
    with pytest.raises(ConnectionError):
        cli.connect()
    assert config.base_url == 'https://xyz.local'

def test_username_and_password_from_environment():
    cli = CLI()
    cli.parse_args(
        ['awx'],
        env={
            'TOWER_USERNAME': 'mary',
            'TOWER_PASSWORD': 'secret'
        }
    )
    with pytest.raises(ConnectionError):
        cli.connect()

    assert config.credentials.default.username == 'mary'
    assert config.credentials.default.password == 'secret'

def test_username_and_password_argv():
    cli = CLI()
    cli.parse_args([
        'awx', '--conf.username', 'mary', '--conf.password', 'secret'
    ])
    with pytest.raises(ConnectionError):
        cli.connect()

    assert config.credentials.default.username == 'mary'
    assert config.credentials.default.password == 'secret'

def test_config_precedence():
    cli = CLI()
    cli.parse_args(
        [
            'awx', '--conf.username', 'mary', '--conf.password', 'secret'
        ],
        env={
            'TOWER_USERNAME': 'IGNORE',
            'TOWER_PASSWORD': 'IGNORE'
        }
    )
    with pytest.raises(ConnectionError):
        cli.connect()

    assert config.credentials.default.username == 'mary'
    assert config.credentials.default.password == 'secret'



from awxkit.utils import PseudoNamespace, random_title, suppress, update_payload
from awxkit.api.resources import resources
from awxkit.api.mixins import HasCreate
import awxkit.exceptions as exc
from . import base
from . import page


class InstanceGroup(HasCreate, base.Base):

    def add_instance(self, instance):
        with suppress(exc.NoContent):
            self.related.instances.post(dict(id=instance.id))

    def remove_instance(self, instance):
        with suppress(exc.NoContent):
            self.related.instances.post(dict(id=instance.id, disassociate=True))

    def payload(self, **kwargs):
        payload = PseudoNamespace(name=kwargs.get('name') or
                                  'Instance Group - {}'.format(random_title()))
        fields = ('policy_instance_percentage', 'policy_instance_minimum', 'policy_instance_list')
        update_payload(payload, fields, kwargs)
        return payload

    def create_payload(self, name='', **kwargs):
        payload = self.payload(name=name, **kwargs)
        return payload

    def create(self, name='', **kwargs):
        payload = self.create_payload(name=name, **kwargs)
        return self.update_identity(InstanceGroups(self.connection).post(payload))


page.register_page([resources.instance_group,
                    (resources.instance_groups, 'post')], InstanceGroup)


class InstanceGroups(page.PageList, InstanceGroup):

    pass


page.register_page([resources.instance_groups,
                    resources.related_instance_groups], InstanceGroups)



from __future__ import print_function

from argparse import ArgumentParser
import os
import sys
import threading

_color = threading.local()
_color.enabled = True


__all__ = ['CustomRegistryMeta', 'HelpfulArgumentParser', 'disable_color',
           'color_enabled', 'colored', 'cprint', 'STATUS_COLORS']


STATUS_COLORS = {
    'new': 'grey',
    'pending': 'grey',
    'running': 'yellow',
    'successful': 'green',
    'failed': 'red',
    'error': 'red',
    'canceled': 'grey',
}


class CustomRegistryMeta(type):

    @property
    def registry(cls):
        return dict(
            (command.name, command)
            for command in cls.__subclasses__()
        )


class HelpfulArgumentParser(ArgumentParser):

    def error(self, message):  # pragma: nocover
        """Prints a usage message incorporating the message to stderr and
        exits.
        If you override this in a subclass, it should not return -- it
        should either exit or raise an exception.
        """
        self.print_help(sys.stderr)
        self._print_message('\n')
        self.exit(2, '%s: %s\n' % (self.prog, message))

    def _parse_known_args(self, args, ns):
        for arg in ('-h', '--help'):
            # the -h argument is extraneous; if you leave it off,
            # awx-cli will just print usage info
            if arg in args:
                args.remove(arg)
        return super(HelpfulArgumentParser, self)._parse_known_args(args, ns)


def color_enabled():
    return _color.enabled


def disable_color():
    _color.enabled = False


COLORS = dict(
    list(
        zip(
            [
                'grey', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan',
                'white',
            ],
            list(range(30, 38))
        )
    )
)


def colored(text, color=None):
    '''Colorize text w/ ANSI color sequences'''
    if _color.enabled and os.getenv('ANSI_COLORS_DISABLED') is None:
        fmt_str = '\033[%dm%s'
        if color is not None:
            text = fmt_str % (COLORS[color], text)
        text += '\033[0m'
    return text


def cprint(text, color, **kwargs):
    if _color.enabled:
        print(colored(text, color), **kwargs)
    else:
        print(text, **kwargs)



import logging

from awxkit.api.resources import resources

from . import base
from . import page


log = logging.getLogger(__name__)


class Role(base.Base):

    NATURAL_KEY = ('name',)

    def get_natural_key(self, cache=None):
        if cache is None:
            cache = page.PageCache()

        natural_key = super(Role, self).get_natural_key(cache=cache)
        related_objs = [
            related for name, related in self.related.items()
            if name not in ('users', 'teams')
        ]
        if related_objs:
            related_endpoint = cache.get_page(related_objs[0])
            if related_endpoint is None:
                log.error("Unable to obtain content_object %s for role %s",
                          related_objs[0], self.endpoint)
                return None
            natural_key['content_object'] = related_endpoint.get_natural_key(cache=cache)

        return natural_key


page.register_page(resources.role, Role)


class Roles(page.PageList, Role):

    pass


page.register_page([resources.roles,
                    resources.related_roles,
                    resources.related_object_roles], Roles)



from awxkit.api.resources import resources
from . import base
from . import page


class Instance(base.Base):

    pass


page.register_page(resources.instance, Instance)


class Instances(page.PageList, Instance):

    pass


page.register_page([resources.instances,
                    resources.related_instances], Instances)



from collections import defaultdict
import logging
import re

from awxkit.utils import is_list_or_tuple, not_provided

log = logging.getLogger(__name__)


class URLRegistry(object):

    def __init__(self):
        self.store = defaultdict(dict)
        self.default = {}

    def url_pattern(self, pattern_str):
        """Converts some regex-friendly url pattern (Resources().resource string)
        to a compiled pattern.
        """
        # should account for any relative endpoint w/ query parameters
        pattern = r'^' + pattern_str + r'(\?.*)*$'
        return re.compile(pattern)

    def _generate_url_iterable(self, url_iterable):
        parsed_urls = []
        for url in url_iterable:
            method = not_provided
            if is_list_or_tuple(url):
                url, method = url
            if not is_list_or_tuple(method):
                methods = (method,)
            else:
                methods = method
            for method in methods:
                method_pattern = re.compile(method)
                url_pattern = self.url_pattern(url)
                parsed_urls.append((url_pattern, method_pattern))
        return parsed_urls

    def register(self, *args):
        """Registers a single resource (generic python type or object) to either
        1. a single url string (internally coverted via URLRegistry.url_pattern) and optional method or method iterable
        2. a list or tuple of url string and optional method or method iterables
        for retrieval via get().

        reg.register('/some/path/', ResourceOne)
        reg.get('/some/path/')
        -> ResourceOne
        reg.register('/some/other/path/', 'method', ResourceTwo)
        reg.get('/some/other/path/', 'method')
        -> ResourceTwo
        reg.register('/some/additional/path/', ('method_one', 'method_two'), ResourceThree)
        reg.get('/some/additional/path/', 'method_one')
        -> ResourceThree
        reg.get('/some/additional/path/', 'method_two')
        -> ResourceThree
        reg.register(('/some/new/path/one/', '/some/new/path/two/',
                      ('/some/other/new/path', 'method'),
                      ('/some/other/additional/path/, ('method_one', 'method_two')), ResourceFour))
        reg.get('/some/other/new/path/', 'method')
        -> ResourceFour
        """
        if not args or len(args) == 1:
            raise TypeError('register needs at least a url and Resource.')
        elif len(args) not in (2, 3):
            raise TypeError('register takes at most 3 arguments ({} given).'.format(len(args)))

        if len(args) == 3:  # url, method (iterable), and Resource
            url_iterable = (args[:2],)
            resource = args[2]
        else:
            urls, resource = args
            if not is_list_or_tuple(urls):
                url_iterable = [(urls, not_provided)]
            else:
                url_iterable = urls

        url_iterable = self._generate_url_iterable(url_iterable)
        for url_pattern, method_pattern in url_iterable:
            if url_pattern in self.store and method_pattern in self.store[url_pattern]:
                if method_pattern.pattern == not_provided:
                    exc_msg = '"{0.pattern}" already has methodless registration.'.format(url_pattern)
                else:
                    exc_msg = ('"{0.pattern}" already has registered method "{1.pattern}"'
                               .format(url_pattern, method_pattern))
                raise TypeError(exc_msg)
            self.store[url_pattern][method_pattern] = resource

    def setdefault(self, *args):
        """Establishes a default return value for get() by optional method (iterable).

        reg.setdefault(ResourceOne)
        reg.get('/some/unregistered/path')
        -> ResourceOne
        reg.setdefault('method', ResourceTwo)
        reg.get('/some/registered/methodless/path/', 'method')
        -> ResourceTwo
        reg.setdefault(('method_one', 'method_two'), ResourceThree)
        reg.get('/some/unregistered/path', 'method_two')
        -> ResourceThree
        reg.setdefault('supports.*regex', ResourceFour)
        reg.get('supports123regex')
        -> ResourceFour
        """
        if not args:
            raise TypeError('setdefault needs at least a Resource.')
        if len(args) == 1:  # all methods
            self.default[re.compile('.*')] = args[0]
        elif len(args) == 2:
            if is_list_or_tuple(args[0]):
                methods = args[0]
            else:
                methods = (args[0],)
            for method in methods:
                method_pattern = re.compile(method)
                self.default[method_pattern] = args[1]
        else:
            raise TypeError('setdefault takes at most 2 arguments ({} given).'.format(len(args)))

    def get(self, url, method=not_provided):
        """Returns a single resource by previously registered path and optional method where
        1.  If a registration was methodless and a method is provided to get() the return value will be
            None or, if applicable, a registry default (see setdefault()).
        2.  If a registration included a method (excluding the method wildcard '.*') and no method is provided to get()
            the return value will be None or, if applicable, a registry default.

        reg.register('/some/path/', ResourceOne)
        reg.get('/some/path/')
        -> ResourceOne
        reg.get('/some/path/', 'method')
        -> None
        reg.register('/some/other/path/', 'method', ResourceTwo)
        reg.get('/some/other/path/', 'method')
        -> ResourceTwo
        reg.get('/some/other/path')
        -> None
        reg.register('/some/additional/path/', '.*', ResourceThree)
        reg.get('/some/additional/path/', 'method')
        -> ResourceThree
        reg.get('/some/additional/path/')
        -> ResourceThree
        """
        registered_type = None
        default_methods = list(self.default)
        # Make sure dot character evaluated last
        default_methods.sort(key=lambda x: x.pattern == '.*')
        for method_key in default_methods:
            if method_key.match(method):
                registered_type = self.default[method_key]
                break

        for re_key in self.store:
            if re_key.match(url):
                keys = list(self.store[re_key])
                keys.sort(key=lambda x: x.pattern == '.*')
                for method_key in keys:
                    if method_key.match(method):
                        registered_type = self.store[re_key][method_key]
                        break
        log.debug('Retrieved {} by url: {}'.format(registered_type, url))
        return registered_type
